#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from octobot_commons.logging.logging_util import get_logger
from octobot_services.api.service_feeds import create_service_feed_factory, start_service_feed


class ServiceFeedFactory:
    """EvaluatorFactory class:
    - Create service feeds
    """

    def __init__(self, octobot):
        self.octobot = octobot

        # Logger
        self.logger = get_logger(self.__class__.__name__)

        self.service_feeds = []

    async def initialize(self):
        in_backtesting = False
        try:
            from octobot_backtesting.api.backtesting import is_backtesting_enabled
            in_backtesting = is_backtesting_enabled(self.octobot.config)
        except ImportError:
            # If can't import octobot_backtesting, this session can't be a backtesting one, nothing to do
            pass
        service_feed_factory = create_service_feed_factory(self.octobot.config,
                                                           self.octobot.async_loop,
                                                           self.octobot.bot_id)
        self.service_feeds = [service_feed_factory.create_service_feed(feed)
                              for feed in service_feed_factory.get_available_service_feeds(in_backtesting)]

    async def create(self):
        for feed in self.service_feeds:
            if not await start_service_feed(feed, False):
                self.logger.error(f"Failed to start {feed.get_name()}. Evaluators requiring this service feed "
                                  f"might not work properly")



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import ccxt
import pytest

from config import CONFIG_ENABLED_OPTION, CONFIG_TRADER
from octobot_trading.exchanges.exchange_manager import ExchangeManager
from tests.test_utils.config import load_test_config
from trading.trader.trader import Trader
from trading.trader.trader_simulator import TraderSimulator
from trading.trader.order import BuyLimitOrder, StopLossOrder, StopLossLimitOrder


# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


class TestOrdersManagers:
    @staticmethod
    async def init_default():
        config = load_test_config()
        config[CONFIG_TRADER][CONFIG_ENABLED_OPTION] = True
        exchange_manager = ExchangeManager(config, ccxt.binance, is_simulated=True)
        await exchange_manager.initialize()
        exchange_inst = exchange_manager.get_exchange()
        trader_real_inst = Trader(config, exchange_inst, 1)
        trader_simulator_inst = TraderSimulator(config, exchange_inst, 1)
        order_manager_inst = trader_real_inst.get_order_manager()
        order_manager_simulator_inst = trader_simulator_inst.get_order_manager()
        return config, exchange_inst, trader_real_inst, order_manager_inst, trader_simulator_inst, \
            order_manager_simulator_inst

    @staticmethod
    def stop(trader):
        trader.stop_order_manager()

    async def test_add_remove_order_to_list(self):
        _, _, trader_inst, order_manager_inst, trader_simulator_inst, \
            order_manager_simulator_inst = await self.init_default()

        # real trader

        assert len(order_manager_inst.get_open_orders()) == 0

        r_order1 = BuyLimitOrder(trader_inst)
        r_order1.order_id = 1
        r_order2 = StopLossOrder(trader_inst)
        r_order1.add_linked_order(r_order2)
        r_order2.add_linked_order(r_order1)
        # add both
        order_manager_inst.add_order_to_list(r_order1)
        order_manager_inst.add_order_to_list(r_order2)
        assert len(order_manager_inst.get_open_orders()) == 2

        r_order3 = BuyLimitOrder(trader_inst)
        r_order3.order_id = 3
        r_order4 = StopLossLimitOrder(trader_inst)
        r_order3.add_linked_order(r_order4)
        r_order4.add_linked_order(r_order3)
        # add both
        order_manager_inst.add_order_to_list(r_order3)
        order_manager_inst.add_order_to_list(r_order4)
        assert len(order_manager_inst.get_open_orders()) == 4

        # add none
        order_manager_inst.add_order_to_list(r_order3)
        order_manager_inst.add_order_to_list(r_order4)
        order_manager_inst.add_order_to_list(r_order1)
        order_manager_inst.add_order_to_list(r_order2)
        assert len(order_manager_inst.get_open_orders()) == 4

        # now remove
        assert order_manager_inst._already_has_real_or_linked_order(r_order2)
        assert order_manager_inst._order_in_list(r_order2)
        order_manager_inst.remove_order_from_list(r_order2)
        assert not order_manager_inst._already_has_real_or_linked_order(r_order2)
        assert not order_manager_inst._order_in_list(r_order2)
        assert len(order_manager_inst.get_open_orders()) == 3
        assert order_manager_inst._already_has_real_or_linked_order(r_order1)
        order_manager_inst.remove_order_from_list(r_order1)
        assert not order_manager_inst._already_has_real_or_linked_order(r_order1)
        assert len(order_manager_inst.get_open_orders()) == 2

        assert order_manager_inst._already_has_real_or_linked_order(r_order3)
        order_manager_inst.remove_order_from_list(r_order3)
        assert not order_manager_inst._already_has_real_or_linked_order(r_order3)
        assert len(order_manager_inst.get_open_orders()) == 1
        assert order_manager_inst._already_has_real_or_linked_order(r_order4)
        assert order_manager_inst._order_in_list(r_order4)
        order_manager_inst.remove_order_from_list(r_order4)
        assert not order_manager_inst._already_has_real_or_linked_order(r_order4)
        assert not order_manager_inst._order_in_list(r_order4)
        assert len(order_manager_inst.get_open_orders()) == 0

        # simulated trader

        assert len(order_manager_simulator_inst.get_open_orders()) == 0

        r_order11 = BuyLimitOrder(trader_simulator_inst)
        r_order12 = StopLossOrder(trader_simulator_inst)
        r_order11.add_linked_order(r_order12)
        r_order12.add_linked_order(r_order11)
        # add both
        order_manager_simulator_inst.add_order_to_list(r_order11)
        order_manager_simulator_inst.add_order_to_list(r_order12)
        assert len(order_manager_simulator_inst.get_open_orders()) == 2

        r_order13 = BuyLimitOrder(trader_simulator_inst)
        r_order14 = StopLossLimitOrder(trader_simulator_inst)
        r_order13.add_linked_order(r_order14)
        r_order14.add_linked_order(r_order13)
        # add both
        order_manager_simulator_inst.add_order_to_list(r_order13)
        order_manager_simulator_inst.add_order_to_list(r_order14)
        assert len(order_manager_simulator_inst.get_open_orders()) == 4

        # add none
        order_manager_simulator_inst.add_order_to_list(r_order13)
        order_manager_simulator_inst.add_order_to_list(r_order14)
        order_manager_simulator_inst.add_order_to_list(r_order11)
        order_manager_simulator_inst.add_order_to_list(r_order12)
        assert len(order_manager_simulator_inst.get_open_orders()) == 4

        # now remove

        assert not order_manager_inst._order_in_list(r_order13)
        assert not order_manager_inst._order_in_list(r_order14)
        assert not order_manager_inst._order_in_list(r_order11)
        assert not order_manager_inst._order_in_list(r_order12)

        assert order_manager_simulator_inst._order_in_list(r_order11)
        order_manager_simulator_inst.remove_order_from_list(r_order11)
        assert len(order_manager_simulator_inst.get_open_orders()) == 3
        assert not order_manager_simulator_inst._order_in_list(r_order11)
        order_manager_simulator_inst.remove_order_from_list(r_order12)
        assert len(order_manager_simulator_inst.get_open_orders()) == 2
        order_manager_simulator_inst.remove_order_from_list(r_order14)
        assert len(order_manager_simulator_inst.get_open_orders()) == 1
        assert order_manager_simulator_inst._order_in_list(r_order13)
        order_manager_simulator_inst.remove_order_from_list(r_order13)
        assert len(order_manager_simulator_inst.get_open_orders()) == 0
        assert not order_manager_simulator_inst._order_in_list(r_order13)

    async def test_update_last_symbol_list(self):
        pass

    async def test_update_last_symbol_prices(self):
        pass

    async def test_in_run(self):
        await self.init_default()



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from services.Dispatchers.dispatcher_creator import DispatcherCreator
from tests.test_utils.config import load_test_config


def test_create_dispatchers(event_loop):
    config = load_test_config()
    dispatchers = DispatcherCreator.create_dispatchers(config, event_loop)
    assert len(dispatchers) == 3  # creates all dispatchers but none is ready (no config for associated services)
    for dispatcher in dispatchers:
        assert not dispatcher.get_is_setup_correctly()



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from octobot_backtesting.api.backtesting import is_backtesting_enabled
from octobot_tentacles_manager.api.configurator import get_tentacles_setup_config
from octobot.community.community_manager import CommunityManager


class Initializer:
    """Initializer class:
    - Initialize services, constants and tools
    """

    def __init__(self, octobot):
        self.octobot = octobot

    async def create(self):
        # initialize tentacle configuration
        self.octobot.tentacles_setup_config = await get_tentacles_setup_config()
        # initialize tools
        self._init_metrics()

    def _init_metrics(self):
        if not is_backtesting_enabled(self.octobot.config):
            self.octobot.community_handler = CommunityManager(self.octobot.octobot_api)



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from math import nan

from octobot_commons.dict_util import get_value_or_default


def test_get_value_or_default():
    test_dict = {"a": 1, "b": 2, "c": 3}
    assert get_value_or_default(test_dict, "b", default="") == 2
    assert get_value_or_default(test_dict, "d", nan) is nan
    assert get_value_or_default(test_dict, "d", default="") == ""
    assert get_value_or_default(test_dict, "z") is None



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import ccxt
import pytest

from config import TraderOrderType
from tests.test_utils.config import load_test_config
from tests.test_utils.order_util import fill_market_order, fill_limit_or_stop_order
from octobot_trading.exchanges import ExchangeManager
from trading.trader.order import BuyMarketOrder, SellLimitOrder, BuyLimitOrder, SellMarketOrder
from trading.trader.portfolio import Portfolio
from trading.trader.sub_portfolio import SubPortfolio
from trading.trader.trader_simulator import TraderSimulator


# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


class TestSubPortfolio:
    DEFAULT_PERCENT = 0.4

    @staticmethod
    async def init_default():
        config = load_test_config()
        exchange_manager = ExchangeManager(config, ccxt.binance, is_simulated=True)
        await exchange_manager.initialize()
        exchange_inst = exchange_manager.get_exchange()
        trader_inst = TraderSimulator(config, exchange_inst, 1)
        portfolio_inst = Portfolio(config, trader_inst)
        await portfolio_inst.initialize()
        trader_inst.stop_order_manager()
        sub_portfolio_inst = SubPortfolio(config, trader_inst, portfolio_inst, TestSubPortfolio.DEFAULT_PERCENT)
        await sub_portfolio_inst.initialize()
        return config, portfolio_inst, exchange_inst, trader_inst, sub_portfolio_inst

    async def test_load_portfolio(self):
        _, _, _, _, sub_portfolio_inst = await self.init_default()
        await sub_portfolio_inst._load_portfolio()
        assert sub_portfolio_inst.portfolio == {'BTC': {'available': 10 * self.DEFAULT_PERCENT,
                                                        'total': 10 * self.DEFAULT_PERCENT},
                                                'USD': {'available': 1000 * self.DEFAULT_PERCENT,
                                                        'total': 1000 * self.DEFAULT_PERCENT}
                                                }

    async def test_get_currency_portfolio(self):
        _, _, _, _, sub_portfolio_inst = await self.init_default()
        assert sub_portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == 10 * self.DEFAULT_PERCENT
        assert sub_portfolio_inst.get_currency_portfolio("BTC", Portfolio.TOTAL) == 10 * self.DEFAULT_PERCENT
        assert sub_portfolio_inst.get_currency_portfolio("NANO", Portfolio.TOTAL) == 0

    async def test_get_currency_multiple_sub_portfolio(self):
        config, portfolio_inst, _, trader_inst, _ = await self.init_default()

        sub_portfolio_inst_2 = SubPortfolio(config, trader_inst, portfolio_inst, 0.2)
        await sub_portfolio_inst_2.initialize()
        sub_portfolio_inst_3 = SubPortfolio(config, trader_inst, portfolio_inst, 0.1)
        await sub_portfolio_inst_3.initialize()
        sub_portfolio_inst_4 = SubPortfolio(config, trader_inst, portfolio_inst, 0.7)
        await sub_portfolio_inst_4.initialize()

        assert sub_portfolio_inst_2.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == 10 * 0.2
        assert sub_portfolio_inst_2.get_currency_portfolio("BTC", Portfolio.TOTAL) == 10 * 0.2
        assert sub_portfolio_inst_2.get_currency_portfolio("NANO", Portfolio.TOTAL) == 0

        assert sub_portfolio_inst_3.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == 10 * 0.1
        assert sub_portfolio_inst_3.get_currency_portfolio("BTC", Portfolio.TOTAL) == 10 * 0.1
        assert sub_portfolio_inst_3.get_currency_portfolio("NANO", Portfolio.TOTAL) == 0

        assert sub_portfolio_inst_4.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == 10 * 0.7
        assert sub_portfolio_inst_4.get_currency_portfolio("BTC", Portfolio.TOTAL) == 10 * 0.7
        assert sub_portfolio_inst_4.get_currency_portfolio("NANO", Portfolio.TOTAL) == 0

    async def test_update_portfolio_available(self):
        _, portfolio_inst, _, trader_inst, sub_portfolio_inst = await self.init_default()

        # Test buy order
        market_buy = BuyMarketOrder(trader_inst)
        market_buy.new(TraderOrderType.BUY_MARKET,
                       "BTC/USD",
                       70,
                       10,
                       70)

        # test sub
        # test buy order creation
        sub_portfolio_inst.update_portfolio_available(market_buy, True)
        assert sub_portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == 10 * self.DEFAULT_PERCENT
        assert sub_portfolio_inst.get_currency_portfolio("USD", Portfolio.AVAILABLE) == 1000*self.DEFAULT_PERCENT-700
        assert sub_portfolio_inst.get_currency_portfolio("BTC", Portfolio.TOTAL) == 10 * self.DEFAULT_PERCENT
        assert sub_portfolio_inst.get_currency_portfolio("USD", Portfolio.TOTAL) == 1000 * self.DEFAULT_PERCENT
        # test parent
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == 10
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.AVAILABLE) == 300
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.TOTAL) == 10
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.TOTAL) == 1000

        # test buy order canceled --> return to init state and the update_portfolio will sync TOTAL with AVAILABLE
        sub_portfolio_inst.update_portfolio_available(market_buy, False)
        assert sub_portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == 10 * self.DEFAULT_PERCENT
        assert sub_portfolio_inst.get_currency_portfolio("USD", Portfolio.AVAILABLE) == 1000 * self.DEFAULT_PERCENT
        assert sub_portfolio_inst.get_currency_portfolio("BTC", Portfolio.TOTAL) == 10 * self.DEFAULT_PERCENT
        assert sub_portfolio_inst.get_currency_portfolio("USD", Portfolio.TOTAL) == 1000 * self.DEFAULT_PERCENT

        # test parent
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == 10
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.AVAILABLE) == 1000
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.TOTAL) == 10
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.TOTAL) == 1000

        # Test sell order
        limit_sell = SellLimitOrder(trader_inst)
        limit_sell.new(TraderOrderType.SELL_LIMIT,
                       "BTC/USD",
                       60,
                       8,
                       60)

        # test sub
        # test sell order creation
        sub_portfolio_inst.update_portfolio_available(limit_sell, True)
        assert sub_portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == 10 * self.DEFAULT_PERCENT - 8
        assert sub_portfolio_inst.get_currency_portfolio("USD", Portfolio.AVAILABLE) == 1000 * self.DEFAULT_PERCENT
        assert sub_portfolio_inst.get_currency_portfolio("BTC", Portfolio.TOTAL) == 10 * self.DEFAULT_PERCENT
        assert sub_portfolio_inst.get_currency_portfolio("USD", Portfolio.TOTAL) == 1000 * self.DEFAULT_PERCENT
        # test parent
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == 2
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.AVAILABLE) == 1000
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.TOTAL) == 10
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.TOTAL) == 1000

        # test sell order canceled --> return to init state and the update_portfolio will sync TOTAL with AVAILABLE
        sub_portfolio_inst.update_portfolio_available(limit_sell, False)
        assert sub_portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == 10 * self.DEFAULT_PERCENT
        assert sub_portfolio_inst.get_currency_portfolio("USD", Portfolio.AVAILABLE) == 1000 * self.DEFAULT_PERCENT
        assert sub_portfolio_inst.get_currency_portfolio("BTC", Portfolio.TOTAL) == 10 * self.DEFAULT_PERCENT
        assert sub_portfolio_inst.get_currency_portfolio("USD", Portfolio.TOTAL) == 1000 * self.DEFAULT_PERCENT

        # test parent
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == 10
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.AVAILABLE) == 1000
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.TOTAL) == 10
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.TOTAL) == 1000

    async def test_update_portfolio(self):
        _, portfolio_inst, _, trader_inst, sub_portfolio_inst = await self.init_default()

        # Test buy order
        limit_buy = BuyLimitOrder(trader_inst)
        limit_buy.new(TraderOrderType.BUY_LIMIT,
                      "BTC/USD",
                      70,
                      10,
                      70)

        # update portfolio with creations
        sub_portfolio_inst.update_portfolio_available(limit_buy, True)
        assert sub_portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == 10 * self.DEFAULT_PERCENT
        assert sub_portfolio_inst.get_currency_portfolio("USD", Portfolio.AVAILABLE) == 1000*self.DEFAULT_PERCENT - 700

        # test parent
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == 10
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.AVAILABLE) == 300

        await fill_limit_or_stop_order(limit_buy, 69, 71)

        await sub_portfolio_inst.update_portfolio(limit_buy)
        assert sub_portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == (10 * self.DEFAULT_PERCENT)+10
        assert sub_portfolio_inst.get_currency_portfolio("USD", Portfolio.AVAILABLE) == 1000*self.DEFAULT_PERCENT-700
        assert sub_portfolio_inst.get_currency_portfolio("BTC", Portfolio.TOTAL) == (10 * self.DEFAULT_PERCENT) + 10
        assert sub_portfolio_inst.get_currency_portfolio("USD", Portfolio.TOTAL) == 1000 * self.DEFAULT_PERCENT - 700

        # test parent
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == 20
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.AVAILABLE) == 300
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.TOTAL) == 20
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.TOTAL) == 300

        # Test buy order
        market_sell = SellMarketOrder(trader_inst)
        market_sell.new(TraderOrderType.SELL_MARKET,
                        "BTC/USD",
                        80,
                        8,
                        80)

        # update portfolio with creations
        sub_portfolio_inst.update_portfolio_available(market_sell, True)
        assert sub_portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == 10*self.DEFAULT_PERCENT + 2
        assert sub_portfolio_inst.get_currency_portfolio("USD", Portfolio.AVAILABLE) == 1000*self.DEFAULT_PERCENT-700

        # test parent
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == 12
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.AVAILABLE) == 300

        await fill_market_order(market_sell, 80)

        # when filling market sell
        await sub_portfolio_inst.update_portfolio(market_sell)
        assert sub_portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == 10*self.DEFAULT_PERCENT + 2
        assert sub_portfolio_inst.get_currency_portfolio("USD", Portfolio.AVAILABLE) == 1000*self.DEFAULT_PERCENT-60
        assert sub_portfolio_inst.get_currency_portfolio("BTC", Portfolio.TOTAL) == 10*self.DEFAULT_PERCENT + 2
        assert sub_portfolio_inst.get_currency_portfolio("USD", Portfolio.TOTAL) == 1000*self.DEFAULT_PERCENT-60

        # test parent
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == 12
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.AVAILABLE) == 940
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.TOTAL) == 12
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.TOTAL) == 940

    async def test_reset_portfolio_available(self):
        _, portfolio_inst, _, trader_inst, sub_portfolio_inst = await self.init_default()

        # Test buy order
        limit_sell = SellLimitOrder(trader_inst)
        limit_sell.new(TraderOrderType.SELL_LIMIT,
                       "BTC/USD",
                       90,
                       4,
                       90)

        sub_portfolio_inst.update_portfolio_available(limit_sell, True)
        sub_portfolio_inst.reset_portfolio_available()
        assert sub_portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == 10*self.DEFAULT_PERCENT
        assert sub_portfolio_inst.get_currency_portfolio("USD", Portfolio.AVAILABLE) == 1000*self.DEFAULT_PERCENT
        assert sub_portfolio_inst.get_currency_portfolio("BTC", Portfolio.TOTAL) == 10*self.DEFAULT_PERCENT
        assert sub_portfolio_inst.get_currency_portfolio("USD", Portfolio.TOTAL) == 1000*self.DEFAULT_PERCENT

        # test parent
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == 10
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.AVAILABLE) == 1000
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.TOTAL) == 10
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.TOTAL) == 1000



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import ccxt
import pytest

from octobot_trading.exchanges import ExchangeManager
from evaluator.symbol_evaluator import SymbolEvaluator
from trading.trader.trader_simulator import TraderSimulator
from evaluator.cryptocurrency_evaluator import CryptocurrencyEvaluator
from evaluator.evaluator_creator import EvaluatorCreator
from tests.test_utils.config import load_test_config
from octobot_commons.tentacles_management.advanced_manager import AdvancedManager
from trading.trader.portfolio import Portfolio
from evaluator.Updaters.global_price_updater import GlobalPriceUpdater
from evaluator.evaluator_task_manager import EvaluatorTaskManager
from config import TimeFrames
from trading.util.trading_config_util import get_activated_trading_mode

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def _get_tools(event_loop):
    symbol = "BTC/USDT"
    exchange_traders = {}
    exchange_traders2 = {}
    config = load_test_config()
    time_frame = TimeFrames.ONE_HOUR
    AdvancedManager.create_class_list(config)
    exchange_manager = ExchangeManager(config, ccxt.binance, is_simulated=True)
    await exchange_manager.initialize()
    exchange_inst = exchange_manager.get_exchange()
    global_price_updater = GlobalPriceUpdater(exchange_inst)
    trader_inst = TraderSimulator(config, exchange_inst, 0.3)
    trader_inst.stop_order_manager()
    trader_inst2 = TraderSimulator(config, exchange_inst, 0.3)
    trader_inst2.stop_order_manager()
    crypto_currency_evaluator = CryptocurrencyEvaluator(config, "Bitcoin", [])
    symbol_evaluator = SymbolEvaluator(config, symbol, crypto_currency_evaluator)
    exchange_traders[exchange_inst.get_name()] = trader_inst
    exchange_traders2[exchange_inst.get_name()] = trader_inst2
    symbol_evaluator.set_trader_simulators(exchange_traders)
    symbol_evaluator.set_traders(exchange_traders2)
    symbol_evaluator.strategies_eval_lists[exchange_inst.get_name()] = \
        EvaluatorCreator.create_strategies_eval_list(config)
    trading_mode_inst = get_activated_trading_mode(config)(config, exchange_inst)
    evaluator_task_manager = EvaluatorTaskManager(config, time_frame, global_price_updater,
                                                  symbol_evaluator, exchange_inst, trading_mode_inst, [],
                                                  event_loop)
    trader_inst.portfolio.portfolio["USDT"] = {
        Portfolio.TOTAL: 2000,
        Portfolio.AVAILABLE: 2000
    }
    return symbol_evaluator, exchange_inst, time_frame, evaluator_task_manager, symbol


async def test_init(event_loop):
    symbol_evaluator, exchange_inst, time_frame, evaluator_task_manager, _ = await _get_tools(event_loop)
    assert symbol_evaluator.trading_mode_instances[exchange_inst.get_name()]
    assert symbol_evaluator.evaluator_task_managers[exchange_inst.get_name()][time_frame] == \
        evaluator_task_manager



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from asyncio import CancelledError
from threading import Thread

import pytest

from backtesting.backtesting_util import start_bot
from core.octobot import OctoBot
from tests.test_utils.config import load_test_config


def stop_bot(bot):
    thread = Thread(target=bot.stop)
    thread.start()
    thread.join()


async def create_bot() -> OctoBot:
    # launch a bot
    config = load_test_config()
    return OctoBot(config)


async def initialize_bot(bot):
    await bot.initialize()


async def call_stop_later(time, event_loop, bot):
    event_loop.call_later(time, stop_bot, bot)


async def start_bot_with_raise(bot, run_in_new_thread=False):
    with pytest.raises(CancelledError):
        await start_bot(bot, run_in_new_thread)



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from random import randint

from config import TimeFrames, PriceIndexes
from octobot_trading.exchanges import SymbolData


class TestExchangeSymbolData:
    def test_update_symbol_candles(self):
        symbol = "BTC/USDT"
        tf_1 = TimeFrames.FIVE_MINUTES
        tf_2 = TimeFrames.ONE_HOUR
        tf_3 = TimeFrames.FOUR_HOURS
        test_inst = SymbolData(symbol)

        fake_candle_1 = self.create_fake_candle(10)
        fake_candle_2 = self.create_fake_candle(11)
        test_inst.update_symbol_candles(tf_1, [fake_candle_1, fake_candle_2], replace_all=True)

        # should be empty
        assert not test_inst.get_candle_data(tf_2)
        assert not test_inst.get_candle_data(tf_3)
        # shouldn't be empty
        assert test_inst.get_candle_data(tf_1)

        candle_data = test_inst.get_candle_data(tf_1)
        # test values
        assert candle_data.get_symbol_close_candles()[-1] == fake_candle_2[PriceIndexes.IND_PRICE_CLOSE.value]
        assert candle_data.get_symbol_open_candles()[-2] == fake_candle_1[PriceIndexes.IND_PRICE_OPEN.value]

        # test update last candle
        fake_candle_3 = self.create_fake_candle(11)
        test_inst.update_symbol_candles(tf_1, fake_candle_3, replace_all=False)
        candle_data = test_inst.get_candle_data(tf_1)
        assert candle_data.get_symbol_close_candles()[-1] == fake_candle_3[PriceIndexes.IND_PRICE_CLOSE.value]
        assert candle_data.get_symbol_high_candles()[-2] == fake_candle_1[PriceIndexes.IND_PRICE_HIGH.value]

        # test new candle
        fake_candle_4 = self.create_fake_candle(20)
        test_inst.update_symbol_candles(tf_1, fake_candle_4, replace_all=False)
        candle_data = test_inst.get_candle_data(tf_1)
        assert candle_data.get_symbol_close_candles()[-1] == fake_candle_4[PriceIndexes.IND_PRICE_CLOSE.value]
        assert candle_data.get_symbol_time_candles()[-1] == fake_candle_4[PriceIndexes.IND_PRICE_TIME.value]
        assert candle_data.get_symbol_volume_candles()[-1] == fake_candle_4[PriceIndexes.IND_PRICE_VOL.value]
        assert candle_data.get_symbol_open_candles()[-1] == fake_candle_4[PriceIndexes.IND_PRICE_OPEN.value]

        # test limit
        fake_candle_5 = self.create_fake_candle(30)
        test_inst.update_symbol_candles(tf_1, fake_candle_5, replace_all=False)
        candle_data = test_inst.get_candle_data(tf_1)
        assert len(candle_data.get_symbol_volume_candles(limit=3)) == 3

        # test lists
        fake_candle_6 = self.create_fake_candle(12)
        test_inst.update_symbol_candles(tf_2, fake_candle_6, replace_all=False)
        fake_candle_7 = self.create_fake_candle(13)
        test_inst.update_symbol_candles(tf_2, fake_candle_7, replace_all=False)
        fake_candle_8 = self.create_fake_candle(12)
        test_inst.update_symbol_candles(tf_3, fake_candle_8, replace_all=True)

        tf2_data = test_inst.get_candle_data(tf_2).get_symbol_close_candles(return_list=True)
        assert tf2_data[-1] == fake_candle_7[PriceIndexes.IND_PRICE_CLOSE.value]
        assert tf2_data[-2] == fake_candle_6[PriceIndexes.IND_PRICE_CLOSE.value]

        tf3_data = test_inst.get_candle_data(tf_3).get_symbol_time_candles(return_list=True)
        assert tf3_data[-1] == fake_candle_8[PriceIndexes.IND_PRICE_TIME.value]

        tf1_data_close = test_inst.get_candle_data(tf_1).get_symbol_close_candles(return_list=True)
        tf1_data_vol = test_inst.get_candle_data(tf_1).get_symbol_volume_candles(return_list=True)
        assert tf1_data_close[-1] == fake_candle_5[PriceIndexes.IND_PRICE_CLOSE.value]
        assert tf1_data_vol[-1] == fake_candle_5[PriceIndexes.IND_PRICE_VOL.value]

        # test replace
        fake_candle_9 = self.create_fake_candle(40)
        test_inst.update_symbol_candles(tf_1, fake_candle_9, replace_all=True)
        tf1_data_low = test_inst.get_candle_data(tf_1).get_symbol_low_candles(return_list=True)
        assert tf1_data_low[-1] == fake_candle_9[PriceIndexes.IND_PRICE_LOW.value]

    def test_max_symbol_candles(self):
        symbol = "BTC/USDT"
        tf = TimeFrames.ONE_DAY
        test_inst = SymbolData(symbol)

        fake_candles = [self.create_fake_candle(i) for i in range(SymbolData.MAX_CANDLES_COUNT)]
        test_inst.update_symbol_candles(tf, fake_candles, replace_all=True)
        assert len(test_inst.get_candle_data(tf).get_symbol_volume_candles()) == SymbolData.MAX_CANDLES_COUNT

        # test to add new one
        fake_candle = self.create_fake_candle(1000)
        test_inst.update_symbol_candles(tf, fake_candle, replace_all=False)
        candle_data = test_inst.get_candle_data(tf)
        assert len(candle_data.get_symbol_volume_candles()) == SymbolData.MAX_CANDLES_COUNT
        assert candle_data.get_symbol_close_candles()[-1] == fake_candle[PriceIndexes.IND_PRICE_CLOSE.value]
        assert candle_data.get_symbol_time_candles()[-1] == fake_candle[PriceIndexes.IND_PRICE_TIME.value]
        assert candle_data.get_symbol_volume_candles()[-1] == fake_candle[PriceIndexes.IND_PRICE_VOL.value]
        assert candle_data.get_symbol_open_candles()[-1] == fake_candle[PriceIndexes.IND_PRICE_OPEN.value]

    @staticmethod
    def create_fake_candle(time):
        return [
            time,
            randint(0, 100),  # open
            randint(0, 100),  # high
            randint(0, 100),  # low
            randint(0, 100),  # close
            randint(0, 100),  # vol
        ]



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import random
import time
import pytest

import ccxt

from octobot_trading.exchanges import ExchangeManager
from config import TradeOrderSide, SIMULATOR_LAST_PRICES_TO_CHECK, TraderOrderType, OrderStatus
from tests.test_utils.config import load_test_config
from trading.trader.order import Order
from trading.trader.trader_simulator import TraderSimulator


# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


class TestOrder:
    @staticmethod
    async def init_default():
        config = load_test_config()
        exchange_manager = ExchangeManager(config, ccxt.binance, is_simulated=True)
        await exchange_manager.initialize()
        exchange_inst = exchange_manager.get_exchange()
        trader_inst = TraderSimulator(config, exchange_inst, 2)
        order_inst = Order(trader_inst)
        return config, order_inst, trader_inst, exchange_inst

    @staticmethod
    def stop(trader):
        trader.stop_order_manager()

    async def test_get_profitability(self):
        _, _, trader_inst, _ = await self.init_default()

        # Test filled_price > create_last_price
        # test side SELL
        order_filled_sup_side_sell_inst = Order(trader_inst)
        order_filled_sup_side_sell_inst.side = TradeOrderSide.SELL
        order_filled_sup_side_sell_inst.filled_price = 10
        order_filled_sup_side_sell_inst.created_last_price = 9
        assert order_filled_sup_side_sell_inst.get_profitability() == (-(1 - 10 / 9))

        # test side BUY
        order_filled_sup_side_sell_inst = Order(trader_inst)
        order_filled_sup_side_sell_inst.side = TradeOrderSide.BUY
        order_filled_sup_side_sell_inst.filled_price = 15.114778
        order_filled_sup_side_sell_inst.created_last_price = 7.265
        assert order_filled_sup_side_sell_inst.get_profitability() == (1 - 15.114778 / 7.265)

        # Test filled_price < create_last_price
        # test side SELL
        order_filled_sup_side_sell_inst = Order(trader_inst)
        order_filled_sup_side_sell_inst.side = TradeOrderSide.SELL
        order_filled_sup_side_sell_inst.filled_price = 11.556877
        order_filled_sup_side_sell_inst.created_last_price = 20
        assert order_filled_sup_side_sell_inst.get_profitability() == (1 - 20 / 11.556877)

        # test side BUY
        order_filled_sup_side_sell_inst = Order(trader_inst)
        order_filled_sup_side_sell_inst.side = TradeOrderSide.BUY
        order_filled_sup_side_sell_inst.filled_price = 8
        order_filled_sup_side_sell_inst.created_last_price = 14.35
        assert order_filled_sup_side_sell_inst.get_profitability() == (-(1 - 14.35 / 8))

        # Test filled_price == create_last_price
        # test side SELL
        order_filled_sup_side_sell_inst = Order(trader_inst)
        order_filled_sup_side_sell_inst.side = TradeOrderSide.SELL
        order_filled_sup_side_sell_inst.filled_price = 1517374.4567
        order_filled_sup_side_sell_inst.created_last_price = 1517374.4567
        assert order_filled_sup_side_sell_inst.get_profitability() == 0

        # test side BUY
        order_filled_sup_side_sell_inst = Order(trader_inst)
        order_filled_sup_side_sell_inst.side = TradeOrderSide.BUY
        order_filled_sup_side_sell_inst.filled_price = 0.4275587387858527
        order_filled_sup_side_sell_inst.created_last_price = 0.4275587387858527
        assert order_filled_sup_side_sell_inst.get_profitability() == 0

        self.stop(trader_inst)

    async def test_check_last_prices(self):
        _, order_inst, trader_inst, _ = await self.init_default()

        # test price in last trades
        # test inferior TRUE
        max_price = 10
        min_price = 4
        order_inst.creation_time = time.time()
        recent_trades = [{"price": random.uniform(min_price, max_price), "timestamp": time.time()}
                         for _ in range(0, SIMULATOR_LAST_PRICES_TO_CHECK)]

        # append validating trade
        recent_trades.append({"price": min_price, "timestamp": time.time()})
        order_inst.last_prices = recent_trades
        assert order_inst.check_last_prices(max_price, inferior=True)

        # test inferior FALSE
        max_price = 10.454677
        min_price = 2.4273
        order_inst.creation_time = time.time()
        recent_trades = [{"price": random.uniform(min_price, max_price), "timestamp": time.time()}
                         for _ in range(0, SIMULATOR_LAST_PRICES_TO_CHECK)]

        # append validating trade
        recent_trades.append({"price": max_price, "timestamp": time.time()})
        order_inst.last_prices = recent_trades
        assert order_inst.check_last_prices(random.uniform(min_price, max_price - 1), inferior=False)

        # test price not in last trades
        # test inferior TRUE
        max_price = 7456.15555632315
        min_price = 1421.1488845
        order_inst.creation_time = time.time()
        recent_trades = [{"price": random.uniform(min_price, max_price), "timestamp": time.time()}
                         for _ in range(0, SIMULATOR_LAST_PRICES_TO_CHECK)]

        order_inst.last_prices = recent_trades
        assert not order_inst.check_last_prices(min_price, inferior=True)

        # test inferior FALSE
        max_price = 0.0001243753
        min_price = 0.000012557753
        order_inst.creation_time = time.time()
        recent_trades = [{"price": random.uniform(min_price, max_price), "timestamp": time.time()}
                         for _ in range(0, SIMULATOR_LAST_PRICES_TO_CHECK)]

        order_inst.last_prices = recent_trades
        assert not order_inst.check_last_prices(max_price, inferior=False)

        self.stop(trader_inst)

    async def test_new(self):
        config, order_inst, trader_inst, exchange_inst = await self.init_default()

        # with real trader
        order_inst.new(TraderOrderType.BUY_MARKET,
                       "BTC/USDT",
                       10000,
                       1,
                       price=None,
                       stop_price=None,
                       order_notifier=None)

        assert order_inst.get_order_type() == TraderOrderType.BUY_MARKET
        assert order_inst.get_order_symbol() == "BTC/USDT"
        assert order_inst.get_create_last_price() == 10000
        assert order_inst.get_origin_quantity() == 1
        assert order_inst.get_creation_time() != 0
        assert order_inst.get_currency_and_market() == ('BTC', 'USDT')
        assert order_inst.get_side() is None
        assert order_inst.get_status() == OrderStatus.OPEN

        order_inst.new(TraderOrderType.STOP_LOSS_LIMIT,
                       "ETH/BTC",
                       0.1,
                       5.2,
                       price=0.12,
                       stop_price=0.9,
                       order_notifier=None)
        assert order_inst.origin_stop_price == 0.9
        assert order_inst.last_prices is None
        assert order_inst.origin_price == 0.12

        # with simulated trader
        trader_sim_inst = TraderSimulator(config, exchange_inst, 1)
        order_sim_inst = Order(trader_sim_inst)

        order_sim_inst.new(TraderOrderType.SELL_MARKET,
                           "LTC/USDT",
                           100,
                           3.22,
                           price=None,
                           stop_price=None,
                           order_notifier=None)
        assert order_sim_inst.get_status() == OrderStatus.OPEN

        self.stop(trader_inst)
        self.stop(trader_sim_inst)



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import ccxt
import pytest

from octobot_trading.exchanges import ExchangeManager
from config import *
from tests.test_utils.config import load_test_config
from trading.trader.trader_simulator import TraderSimulator
from trading.trader.trade import Trade
from trading.trader.order import SellLimitOrder

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


class TestTradesManager:
    @staticmethod
    async def init_default():
        config = load_test_config()
        exchange_manager = ExchangeManager(config, ccxt.binance, is_simulated=True)
        await exchange_manager.initialize()
        exchange_inst = exchange_manager.get_exchange()
        trader_inst = TraderSimulator(config, exchange_inst, 1)
        await trader_inst.portfolio.initialize()
        trades_manager_inst = trader_inst.get_trades_manager()
        await trades_manager_inst.initialize()
        return config, exchange_inst, trader_inst, trades_manager_inst

    @staticmethod
    def stop(trader):
        trader.stop_order_manager()

    async def test_get_profitability(self):
        _, _, trader_inst, trades_manager_inst = await self.init_default()
        self.stop(trader_inst)

        profitability, profitability_percent, profitability_diff, market_profitability, \
            initial_portfolio_current_profitability = await trades_manager_inst.get_profitability()
        assert market_profitability is None
        assert profitability == profitability_percent == profitability_diff == \
            initial_portfolio_current_profitability == 0

        trades_manager_inst.portfolio_origin_value = 20
        trades_manager_inst.profitability_percent = 100

        # set all market values to 1 if unset
        nb_currencies = 0
        for currency in \
                trades_manager_inst.only_symbol_currency_filter(trades_manager_inst.origin_crypto_currencies_values):
            if trades_manager_inst.origin_crypto_currencies_values[currency] != 0:
                nb_currencies += 1

        # decrease initial value of btc to force market_profitability change:
        # => same as +100% for BTC (ratio = nb_currencies + 1 /  nb_currencies in %)
        expected_market_profitability = ((nb_currencies + 1) / nb_currencies * 100) - 100
        trades_manager_inst.origin_crypto_currencies_values["BTC"] = 0.5

        profitability, profitability_percent, profitability_diff, market_profitability, \
            initial_portfolio_current_profitability = await trades_manager_inst.get_profitability(True)
        assert profitability == -10
        assert profitability_percent == -50
        assert profitability_diff == -150
        assert market_profitability == expected_market_profitability
        assert initial_portfolio_current_profitability == -50

    async def test_get_current_holdings_values(self):
        _, _, trader_inst, trades_manager_inst = await self.init_default()
        self.stop(trader_inst)
        btc_currency = "BTC"
        usd_currency = "USD"
        ada_currency = "ADA"

        usd_price_in_btc = 50
        trades_manager_inst.current_crypto_currencies_values = {
            btc_currency: 1,
            usd_currency: usd_price_in_btc,
            ada_currency: 10
        }
        holdings = await trades_manager_inst.get_current_holdings_values()
        btc_holding = trader_inst.portfolio.portfolio[btc_currency][CONFIG_PORTFOLIO_TOTAL]
        usd_holding = trader_inst.portfolio.portfolio[usd_currency][CONFIG_PORTFOLIO_TOTAL]
        ada_holding = 0
        portfolio = trader_inst.portfolio.portfolio
        if ada_currency not in portfolio:
            portfolio[ada_currency] = {}
        portfolio[ada_currency][CONFIG_PORTFOLIO_TOTAL] = ada_holding
        assert len(holdings) == len(trades_manager_inst.current_crypto_currencies_values)
        assert holdings[btc_currency] == btc_holding * 1    # ref market is btc => btc_holding*1
        assert holdings[usd_currency] == usd_price_in_btc * usd_holding
        assert holdings[ada_currency] == 0

    async def test_add_select_trade_in_history(self):
        _, exchange_inst, trader_inst, trades_manager_inst = await self.init_default()
        self.stop(trader_inst)
        assert len(trades_manager_inst.get_trade_history()) == 0
        symbol = "BTC/USD"
        new_order = SellLimitOrder(trader_inst)
        new_order.new(TraderOrderType.SELL_LIMIT, symbol, 90, 4, 90)
        new_trade = Trade(exchange_inst, new_order)

        # add new trade
        trades_manager_inst.add_new_trade_in_history(new_trade)
        assert len(trades_manager_inst.get_trade_history()) == 1
        assert trades_manager_inst.get_trade_history()[0] == new_trade

        # doesn't add an existing trade
        trades_manager_inst.add_new_trade_in_history(new_trade)
        assert len(trades_manager_inst.get_trade_history()) == 1
        assert trades_manager_inst.get_trade_history()[0] == new_trade

        # select trade
        assert trades_manager_inst.select_trade_history() == trades_manager_inst.get_trade_history()

        new_order2 = SellLimitOrder(trader_inst)
        new_order2.new(TraderOrderType.SELL_LIMIT, "BTC/EUR", 90, 4, 90)
        new_trade2 = Trade(exchange_inst, new_order2)

        trades_manager_inst.add_new_trade_in_history(new_trade2)
        assert len(trades_manager_inst.get_trade_history()) == 2
        assert trades_manager_inst.select_trade_history(symbol) == [new_trade]

    async def test_get_total_paid_fees(self):
        _, exchange_inst, trader_inst, trades_manager_inst = await self.init_default()
        self.stop(trader_inst)
        symbol = "BTC/USD"
        new_order = SellLimitOrder(trader_inst)
        new_order.new(TraderOrderType.SELL_LIMIT, symbol, 90, 4, 90)
        new_order.fee = {
            FeePropertyColumns.COST.value: 100,
            FeePropertyColumns.CURRENCY.value: "BTC"
        }
        new_trade = Trade(exchange_inst, new_order)
        trades_manager_inst.add_new_trade_in_history(new_trade)

        assert trades_manager_inst.get_total_paid_fees() == {"BTC": 100}

        new_order2 = SellLimitOrder(trader_inst)
        new_order2.new(TraderOrderType.SELL_LIMIT, symbol, 90, 4, 90)
        new_order2.fee = {
            FeePropertyColumns.COST.value: 200,
            FeePropertyColumns.CURRENCY.value: "PLOP"
        }
        new_trade2 = Trade(exchange_inst, new_order2)
        trades_manager_inst.add_new_trade_in_history(new_trade2)

        assert trades_manager_inst.get_total_paid_fees() == {"BTC": 100, "PLOP": 200}

        new_order3 = SellLimitOrder(trader_inst)
        new_order3.new(TraderOrderType.SELL_LIMIT, symbol, 90, 4, 90)
        new_order3.fee = {
            FeePropertyColumns.COST.value: 0.01111,
            FeePropertyColumns.CURRENCY.value: "PLOP"
        }
        new_trade3 = Trade(exchange_inst, new_order3)
        trades_manager_inst.add_new_trade_in_history(new_trade3)

        assert trades_manager_inst.get_total_paid_fees() == {"BTC": 100, "PLOP": 200.01111}



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest

from config import CONFIG_CRYPTO_CURRENCIES, CONFIG_CRYPTO_PAIRS, CONFIG_WILDCARD, CONFIG_CRYPTO_QUOTE
from core.octobot import OctoBot
from evaluator.cryptocurrency_evaluator import CryptocurrencyEvaluator
from evaluator.symbol_evaluator import SymbolEvaluator
from tests.test_utils.config import load_test_config

# All test coroutines will be treated as marked.
from octobot_commons.symbol_util import split_symbol

pytestmark = pytest.mark.asyncio


def create_bot_with_config(config):
    return OctoBot(config, ignore_config=True)


def check_symbol_evaluator_list(e_factory, symbol, exchange, exchange_count, total_symbol_count):
    assert len(e_factory.symbol_evaluator_list) == total_symbol_count
    symbol_eval = e_factory.symbol_evaluator_list[symbol]
    assert isinstance(symbol_eval, SymbolEvaluator)
    assert symbol_eval.symbol == symbol
    assert len(symbol_eval.trader_simulators) == exchange_count
    assert exchange in symbol_eval.trader_simulators
    assert len(symbol_eval.traders) == exchange_count
    assert exchange in symbol_eval.traders
    assert len(symbol_eval.trading_mode_instances) == 1


def check_crypto_currency_evaluator_list(e_factory, crypto_currency, symbol, crypto_currency_count, symbol_count):
    assert len(e_factory.crypto_currency_evaluator_list) == crypto_currency_count
    assert crypto_currency in e_factory.crypto_currency_evaluator_list
    crypto_evaluator = e_factory.crypto_currency_evaluator_list[crypto_currency]
    assert isinstance(crypto_evaluator, CryptocurrencyEvaluator)
    assert crypto_evaluator.crypto_currency == crypto_currency
    assert len(crypto_evaluator.symbol_evaluator_list) == symbol_count
    assert symbol in crypto_evaluator.symbol_evaluator_list
    assert crypto_evaluator.symbol_evaluator_list[symbol] is e_factory.symbol_evaluator_list[symbol]


def perform_evaluator_factory_tests(e_factory, crypto_currency, symbol, exchange,
                                    exchange_count, crypto_currency_count, total_symbol_count, symbol_count):
    # symbol_evaluator_list
    check_symbol_evaluator_list(e_factory, symbol, exchange, exchange_count, total_symbol_count)

    # trading mode instance
    btc_trading_mode = e_factory.symbol_evaluator_list[symbol].trading_mode_instances
    assert exchange in btc_trading_mode

    # crypto_currency_evaluator_list
    check_crypto_currency_evaluator_list(e_factory, crypto_currency, symbol, crypto_currency_count, symbol_count)


async def test_evaluator_factory_with_one_symbol():
    crypto_currency = "Bitcoin"
    symbol = "BTC/USDT"
    exchange = "binance"
    config = load_test_config()
    config[CONFIG_CRYPTO_CURRENCIES] = {
        crypto_currency:
            {
                CONFIG_CRYPTO_PAIRS: [symbol]
            }
    }
    bot = create_bot_with_config(config)
    await bot.initialize()

    perform_evaluator_factory_tests(bot.evaluator_factory, crypto_currency, symbol, exchange, 1, 1, 1, 1)


async def test_evaluator_factory_with_three_symbols():
    crypto_currency = "Ethereum"
    symbols = ["ETH/USDT", "ETH/BTC", "ETH/BNB", "ETH/EOS"]
    exchange = "binance"
    config = load_test_config()
    config[CONFIG_CRYPTO_CURRENCIES] = {
        crypto_currency:
            {
                CONFIG_CRYPTO_PAIRS: symbols
            }
    }
    bot = create_bot_with_config(config)
    await bot.initialize()

    # ETH/BNB and ETH/EOS not in symbol list --> 2 of 4
    perform_evaluator_factory_tests(bot.evaluator_factory, crypto_currency, "ETH/USDT", exchange, 1, 1, 2, 2)
    perform_evaluator_factory_tests(bot.evaluator_factory, crypto_currency, "ETH/BTC", exchange, 1, 1, 2, 2)


async def test_evaluator_factory_with_multiple_currencies():
    eth_crypto_currency = "Ethereum"
    eth_symbols = ["ETH/USDT", "ETH/BTC", "ETH/BNB", "ETH/EOS"]
    btc_crypto_currency = "Bitcoin"
    btc_symbols = ["BTC/USDT", "BTC/BNB"]
    exchange = "binance"
    config = load_test_config()
    config[CONFIG_CRYPTO_CURRENCIES] = {
        eth_crypto_currency:
            {
                CONFIG_CRYPTO_PAIRS: eth_symbols
            },
        btc_crypto_currency:
            {
                CONFIG_CRYPTO_PAIRS: btc_symbols
            }
    }
    bot = create_bot_with_config(config)
    await bot.initialize()

    # ETH/BNB and ETH/EOS not in symbol list --> 2 of 4
    perform_evaluator_factory_tests(bot.evaluator_factory, eth_crypto_currency, "ETH/USDT", exchange, 1, 2, 3, 2)
    perform_evaluator_factory_tests(bot.evaluator_factory, eth_crypto_currency, "ETH/BTC", exchange, 1, 2, 3, 2)

    # BTC/BNB not in symbol list --> 1 of 2
    perform_evaluator_factory_tests(bot.evaluator_factory, btc_crypto_currency, "BTC/USDT", exchange, 1, 2, 3, 1)


async def test_evaluator_factory_with_wildcard():
    btc_crypto_currency = "Bitcoin"
    btc_quote = "BTC"
    exchange = "binance"
    config = load_test_config()
    config[CONFIG_CRYPTO_CURRENCIES] = {
        btc_crypto_currency:
            {
                CONFIG_CRYPTO_PAIRS: CONFIG_WILDCARD,
                CONFIG_CRYPTO_QUOTE: btc_quote
            }
    }
    bot = create_bot_with_config(config)
    await bot.initialize()

    symbol_count = len(set([
        symbol if split_symbol(symbol)[1] == btc_quote else None
        for symbol in bot.evaluator_factory.octobot.exchange_factory.exchanges_list['binance'].exchange.client.symbols
    ])) - 1  # -1 for 1 None (set consequence)

    perform_evaluator_factory_tests(bot.evaluator_factory, btc_crypto_currency, "ETH/BTC", exchange,
                                    1, 1, symbol_count, symbol_count)



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from config.config import load_config
from config import CONFIG_EVALUATOR, CONFIG_TRADING_TENTACLES, TimeFrames, CONFIG_TIME_FRAME


TEST_CONFIG_FOLDER = "tests/static"


def init_config_time_frame_for_tests(config):
    result = []
    for time_frame in config[CONFIG_TIME_FRAME]:
        result.append(TimeFrames(time_frame))
    config[CONFIG_TIME_FRAME] = result


def load_test_config():
    config = load_config(f"{TEST_CONFIG_FOLDER}/config.json")
    config[CONFIG_EVALUATOR] = load_config(f"{TEST_CONFIG_FOLDER}/evaluator_config.json", False)
    config[CONFIG_TRADING_TENTACLES] = load_config(f"{TEST_CONFIG_FOLDER}/trading_config.json", False)
    init_config_time_frame_for_tests(config)
    return config



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import ccxt

from octobot_backtesting.api.backtesting import is_backtesting_enabled, get_backtesting_data_files
from octobot_commons.logging.logging_util import get_logger
from octobot_trading.api.exchange import create_exchange_builder
from octobot_trading.api.trader import is_trader_enabled_in_config, is_trader_simulator_enabled_in_config
from octobot_trading.constants import CONFIG_EXCHANGES
from octobot.logger import init_exchange_chan_logger


class ExchangeFactory:
    """
    - Create exchanges and trades according to configured exchanges
    """

    def __init__(self, octobot, ignore_config=False):
        self.octobot = octobot
        self.ignore_config = ignore_config

        # Logger
        self.logger = get_logger(self.__class__.__name__)

        self.exchange_manager = None
        self.exchange_manager_ids = []
        self.exchange_traders = {}
        self.exchange_trader_simulators = {}
        self.exchange_trading_modes = {}
        self.trading_mode = None
        self.previous_trading_state_manager = None
        self.exchanges_list = {}
        self.global_updaters_by_exchange = {}

        self.available_exchanges = ccxt.exchanges

    async def create(self):
        if self.octobot.config[CONFIG_EXCHANGES]:
            for exchange_class_string in self.octobot.config[CONFIG_EXCHANGES]:
                if exchange_class_string in self.available_exchanges:
                    exchange_builder = create_exchange_builder(self.octobot.config, exchange_class_string) \
                        .has_matrix(self.octobot.evaluator_factory.matrix_id) \
                        .use_tentacles_setup_config(self.octobot.tentacles_setup_config) \
                        .is_rest_only()
                    if is_trader_enabled_in_config(self.octobot.config):
                        exchange_builder.is_real()
                    elif is_trader_simulator_enabled_in_config(self.octobot.config):
                        exchange_builder.is_simulated()
                    if is_backtesting_enabled(self.octobot.config):
                        exchange_builder.is_backtesting(get_backtesting_data_files(self.octobot.config))
                    exchange_manager = await exchange_builder.build()
                    await init_exchange_chan_logger(exchange_manager.id)
                    self.exchange_manager_ids.append(exchange_manager.id)
                else:
                    self.logger.error(f"{exchange_class_string} exchange not found")
        else:
            self.logger.error("No exchange in configuration. OctoBot requires at least one exchange "
                              "to read trading data from. You can add exchanges in the configuration section.")



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import ccxt
import pytest

from config import CONFIG_ENABLED_OPTION, CONFIG_BACKTESTING, TimeFrames, HOURS_TO_SECONDS, PriceIndexes, \
    TraderOrderType, ExchangeConstantsMarketPropertyColumns, FeePropertyColumns, CONFIG_SIMULATOR, \
    CONFIG_SIMULATOR_FEES, CONFIG_SIMULATOR_FEES_MAKER, CONFIG_SIMULATOR_FEES_TAKER, BACKTESTING_DATA_OHLCV, \
    BACKTESTING_DATA_TRADES, ExchangeConstantsOrderColumns
from tests.test_utils.config import load_test_config
from octobot_trading.exchanges import ExchangeManager
from trading.trader.trader_simulator import TraderSimulator


# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


class TestExchangeSimulator:
    DEFAULT_SYMBOL = "BTC/USDT"
    DEFAULT_TF = TimeFrames.ONE_HOUR

    @staticmethod
    async def init_default():
        config = load_test_config()
        config[CONFIG_BACKTESTING][CONFIG_ENABLED_OPTION] = True
        exchange_manager = ExchangeManager(config, ccxt.binance, is_simulated=True)
        await exchange_manager.initialize()
        exchange_inst = exchange_manager.get_exchange()
        exchange_simulator = exchange_inst.get_exchange()
        exchange_simulator.init_candles_offset([TimeFrames.ONE_HOUR, TimeFrames.FOUR_HOURS, TimeFrames.ONE_DAY],
                                               TestExchangeSimulator.DEFAULT_SYMBOL)

        trader_inst = TraderSimulator(config, exchange_inst, 1)
        return config, exchange_inst, exchange_simulator, trader_inst

    @staticmethod
    def stop(trader):
        trader.stop_order_manager()

    async def test_multiple_get_symbol_prices(self):
        _, exchange_inst, _, trader_inst = await self.init_default()
        self.stop(trader_inst)

        first_data = await exchange_inst.get_symbol_prices(
            self.DEFAULT_SYMBOL,
            self.DEFAULT_TF,
            return_list=False)

        second_data = await exchange_inst.get_symbol_prices(
            self.DEFAULT_SYMBOL,
            self.DEFAULT_TF,
            return_list=False)

        # different arrays
        assert first_data is not second_data

        # second is first with DEFAULT_TF difference
        assert first_data[PriceIndexes.IND_PRICE_CLOSE.value][0] == second_data[PriceIndexes.IND_PRICE_CLOSE.value][0]
        assert first_data[PriceIndexes.IND_PRICE_TIME.value][0] == second_data[
            PriceIndexes.IND_PRICE_TIME.value][0]

        # end is end -1 with DEFAULT_TF difference
        assert first_data[PriceIndexes.IND_PRICE_CLOSE.value][-1] == second_data[PriceIndexes.IND_PRICE_CLOSE.value][-2]
        assert first_data[PriceIndexes.IND_PRICE_TIME.value][-1] + HOURS_TO_SECONDS == second_data[
            PriceIndexes.IND_PRICE_TIME.value][-1]

    async def test_get_recent_trades(self):
        _, exchange_inst, _, trader_inst = await self.init_default()
        self.stop(trader_inst)

        await exchange_inst.get_recent_trades(self.DEFAULT_SYMBOL)

    async def test_get_all_currencies_price_ticker(self):
        _, exchange_inst, _, trader_inst = await self.init_default()
        self.stop(trader_inst)

        await exchange_inst.get_all_currencies_price_ticker()

    @staticmethod
    def _assert_fee(fee, currency, price, rate, fee_type):
        assert fee[FeePropertyColumns.CURRENCY.value] == currency
        assert fee[FeePropertyColumns.COST.value] == price
        assert fee[FeePropertyColumns.RATE.value] == rate
        assert fee[FeePropertyColumns.TYPE.value] == fee_type

    async def test_get_trade_fee(self):
        _, exchange_inst, _, _ = await self.init_default()

        # force fees
        exchange_inst.config[CONFIG_SIMULATOR][CONFIG_SIMULATOR_FEES] = {
            CONFIG_SIMULATOR_FEES_MAKER: 0.05,
            CONFIG_SIMULATOR_FEES_TAKER: 0.1
        }

        buy_market_fee = exchange_inst.get_trade_fee("BTC/USD", TraderOrderType.BUY_MARKET,
                                                     10, 100, ExchangeConstantsMarketPropertyColumns.TAKER.value)
        self._assert_fee(buy_market_fee, "BTC", 0.01, 0.001, ExchangeConstantsMarketPropertyColumns.TAKER.value)

        sell_market_fee = exchange_inst.get_trade_fee(
            "BTC/USD",  TraderOrderType.SELL_MARKET, 10, 100, ExchangeConstantsMarketPropertyColumns.TAKER.value)
        self._assert_fee(sell_market_fee, "USD", 1, 0.001, ExchangeConstantsMarketPropertyColumns.TAKER.value)

        buy_limit_fee = exchange_inst.get_trade_fee("BTC/USD", TraderOrderType.BUY_LIMIT,
                                                    10, 100, ExchangeConstantsMarketPropertyColumns.MAKER.value)
        self._assert_fee(buy_limit_fee, "BTC", 0.005, 0.0005, ExchangeConstantsMarketPropertyColumns.MAKER.value)

        sell_limit_fee = exchange_inst.get_trade_fee("BTC/USD", TraderOrderType.SELL_LIMIT,
                                                     10, 100, ExchangeConstantsMarketPropertyColumns.TAKER.value)
        self._assert_fee(sell_limit_fee, "USD", 1, 0.001, ExchangeConstantsMarketPropertyColumns.TAKER.value)

    async def test_should_update_data(self):
        _, exchange_inst, exchange_simulator, trader_inst = await self.init_default()

        # first call
        assert exchange_simulator.should_update_data(TimeFrames.ONE_HOUR, self.DEFAULT_SYMBOL)
        assert exchange_simulator.should_update_data(TimeFrames.FOUR_HOURS, self.DEFAULT_SYMBOL)
        assert exchange_simulator.should_update_data(TimeFrames.ONE_DAY, self.DEFAULT_SYMBOL)

        # call get prices
        await exchange_inst.get_symbol_prices(self.DEFAULT_SYMBOL, TimeFrames.ONE_HOUR)
        await exchange_inst.get_symbol_prices(self.DEFAULT_SYMBOL, TimeFrames.FOUR_HOURS)
        await exchange_inst.get_symbol_prices(self.DEFAULT_SYMBOL, TimeFrames.ONE_DAY)

        # call with trader without order
        assert exchange_simulator.should_update_data(TimeFrames.ONE_HOUR, self.DEFAULT_SYMBOL)
        assert not exchange_simulator.should_update_data(TimeFrames.FOUR_HOURS, self.DEFAULT_SYMBOL)
        assert not exchange_simulator.should_update_data(TimeFrames.ONE_DAY, self.DEFAULT_SYMBOL)
        await exchange_inst.get_symbol_prices(self.DEFAULT_SYMBOL, TimeFrames.ONE_HOUR)

        self.stop(trader_inst)

    @staticmethod
    def _get_start_index_for_timeframe(nb_candles, min_limit, timeframe_multiplier):
        return int(nb_candles - (nb_candles-min_limit) / timeframe_multiplier) - 1

    async def test_init_candles_offset(self):
        _, _, exchange_simulator, _ = await self.init_default()

        timeframes = [TimeFrames.THIRTY_MINUTES, TimeFrames.ONE_HOUR, TimeFrames.TWO_HOURS,
                      TimeFrames.FOUR_HOURS, TimeFrames.ONE_DAY]
        exchange_simulator.init_candles_offset(timeframes, self.DEFAULT_SYMBOL)

        offsets = exchange_simulator.time_frames_offset[self.DEFAULT_SYMBOL]
        ohlcv = exchange_simulator.data[self.DEFAULT_SYMBOL][BACKTESTING_DATA_OHLCV]
        assert ohlcv is exchange_simulator.get_ohlcv(self.DEFAULT_SYMBOL)
        nb_candles = len(ohlcv[TimeFrames.THIRTY_MINUTES.value])
        assert offsets[TimeFrames.THIRTY_MINUTES.value] == \
            self._get_start_index_for_timeframe(nb_candles, exchange_simulator.MIN_LIMIT, 1) + 1
        assert offsets[TimeFrames.ONE_HOUR.value] ==  \
            self._get_start_index_for_timeframe(nb_candles, exchange_simulator.MIN_LIMIT, 2)
        assert offsets[TimeFrames.TWO_HOURS.value] ==  \
            self._get_start_index_for_timeframe(nb_candles, exchange_simulator.MIN_LIMIT, 4)
        assert offsets[TimeFrames.FOUR_HOURS.value] ==  \
            self._get_start_index_for_timeframe(nb_candles, exchange_simulator.MIN_LIMIT, 8)
        assert offsets[TimeFrames.ONE_DAY.value] == 244

    async def test_select_trades(self):
        _, _, exchange_simulator, _ = await self.init_default()
        trades = [
            {
                ExchangeConstantsOrderColumns.TIMESTAMP.value: 1549413007896,
                ExchangeConstantsOrderColumns.PRICE.value: "3415.30000"
            },
            {
                ExchangeConstantsOrderColumns.TIMESTAMP.value: 1549413032879,
                ExchangeConstantsOrderColumns.PRICE.value: "3415.30000"
            },
            {
                ExchangeConstantsOrderColumns.TIMESTAMP.value: 1549413032922,
                ExchangeConstantsOrderColumns.PRICE.value: "3415.90000"
            }
        ]
        exchange_simulator.data[self.DEFAULT_SYMBOL][BACKTESTING_DATA_TRADES] = trades
        assert exchange_simulator.select_trades(10, 10, self.DEFAULT_SYMBOL) == []
        assert exchange_simulator.select_trades(1549413007896, -1, self.DEFAULT_SYMBOL) == []
        with pytest.raises(KeyError):
            assert exchange_simulator.select_trades(1549413007.896, 1549413032.879, "ETH/USD") == []
        assert exchange_simulator.select_trades(exchange_simulator.get_uniform_timestamp(1549413007896),
                                                exchange_simulator.get_uniform_timestamp(1549413032879),
                                                self.DEFAULT_SYMBOL) == [
            {
                ExchangeConstantsOrderColumns.TIMESTAMP.value: exchange_simulator.get_uniform_timestamp(1549413007896),
                ExchangeConstantsOrderColumns.PRICE.value: 3415.30000
            },
            {
                ExchangeConstantsOrderColumns.TIMESTAMP.value: exchange_simulator.get_uniform_timestamp(1549413032879),
                ExchangeConstantsOrderColumns.PRICE.value: 3415.30000
            }
        ]
        assert exchange_simulator.select_trades(exchange_simulator.get_uniform_timestamp(1549413007896),
                                                -1, self.DEFAULT_SYMBOL) == [
            {
                ExchangeConstantsOrderColumns.TIMESTAMP.value: exchange_simulator.get_uniform_timestamp(1549413007896),
                ExchangeConstantsOrderColumns.PRICE.value: 3415.30000
            },
            {
                ExchangeConstantsOrderColumns.TIMESTAMP.value: exchange_simulator.get_uniform_timestamp(1549413032879),
                ExchangeConstantsOrderColumns.PRICE.value: 3415.30000
            },
            {
                ExchangeConstantsOrderColumns.TIMESTAMP.value: exchange_simulator.get_uniform_timestamp(1549413032922),
                ExchangeConstantsOrderColumns.PRICE.value: 3415.90000
            }
        ]



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from octobot_commons.logging.logging_util import get_logger
from octobot_evaluators.api.evaluators import initialize_evaluators, create_all_type_evaluators
from octobot_evaluators.api.initialization import create_matrix_channels
from octobot_trading.exchanges.exchanges import Exchanges
from octobot.logger import init_evaluator_chan_logger


class EvaluatorFactory:
    """EvaluatorFactory class:
    - Create evaluators
    """

    def __init__(self, octobot):
        self.octobot = octobot

        # Logger
        self.logger = get_logger(self.__class__.__name__)

        self.matrix_id = None
        self.tentacles_setup_config = None

    async def initialize(self):
        self.tentacles_setup_config = self.octobot.tentacles_setup_config
        self.matrix_id = await initialize_evaluators(self.octobot.config, self.tentacles_setup_config)
        await create_matrix_channels()

    async def create(self):
        for exchange_configuration in Exchanges.instance().get_all_exchanges():
            await create_all_type_evaluators(
                self.octobot.config,
                tentacles_setup_config=self.tentacles_setup_config,
                matrix_id=self.matrix_id,
                exchange_name=exchange_configuration.exchange_name,
                bot_id=self.octobot.bot_id,
                symbols_by_crypto_currencies=exchange_configuration.symbols_by_crypto_currencies,
                symbols=exchange_configuration.symbols,
                time_frames=exchange_configuration.time_frames
            )
        await init_evaluator_chan_logger()



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import time

from config import CONFIG_PORTFOLIO_FREE, CONFIG_PORTFOLIO_USED, CONFIG_PORTFOLIO_TOTAL, OrderStatus
from octobot_trading.exchanges.data.exchange_personal_data import ExchangePersonalData


class TestExchangePersonalData:
    def test_update_portfolio(self):
        test_inst = ExchangePersonalData()

        # without pf
        test_inst.portfolio = {}
        test_inst.update_portfolio("BTC", 10, 7, 3)
        assert test_inst.portfolio["BTC"][CONFIG_PORTFOLIO_FREE] == 7
        assert test_inst.portfolio["BTC"][CONFIG_PORTFOLIO_USED] == 3
        assert test_inst.portfolio["BTC"][CONFIG_PORTFOLIO_TOTAL] == 10

        # with pf
        test_inst.update_portfolio("ETH", 100, 60, 40)
        assert test_inst.portfolio["BTC"][CONFIG_PORTFOLIO_FREE] == 7
        assert test_inst.portfolio["BTC"][CONFIG_PORTFOLIO_USED] == 3
        assert test_inst.portfolio["BTC"][CONFIG_PORTFOLIO_TOTAL] == 10
        assert test_inst.portfolio["ETH"][CONFIG_PORTFOLIO_FREE] == 60
        assert test_inst.portfolio["ETH"][CONFIG_PORTFOLIO_USED] == 40
        assert test_inst.portfolio["ETH"][CONFIG_PORTFOLIO_TOTAL] == 100

    def test_has_order(self):
        test_inst = ExchangePersonalData()

        # without orders
        test_inst.orders = {}
        assert not test_inst.has_order(10)
        test_inst.orders[10] = self.create_fake_order(10, None, None, None)
        assert test_inst.has_order(10)
        assert not test_inst.has_order(153)

    def test_upsert_order(self):
        test_inst = ExchangePersonalData()

        # without orders
        test_inst.orders = {}
        assert not test_inst.has_order(10)
        test_inst.upsert_order(10, None)
        assert test_inst.has_order(10)

    def test_set_orders(self):
        test_inst = ExchangePersonalData()

        # without orders
        test_inst.orders = {}
        assert not test_inst.has_order(10)
        assert not test_inst.has_order(20)
        test_inst.upsert_orders([
            {"id": 15},
            {"id": 10},
            {"id": 20},
            {"id": 15}
        ])
        assert test_inst.get_order(15)
        assert test_inst.has_order(15)
        assert test_inst.has_order(20)
        assert test_inst.has_order(10)
        assert not test_inst.has_order(12)
        assert not test_inst.has_order(30)

    def test_set_more_than_max_count_orders(self):
        test_inst = ExchangePersonalData()
        test_inst.orders = {}
        test_inst._MAX_ORDERS_COUNT = 500
        nb_max_stored_orders = test_inst._MAX_ORDERS_COUNT

        max_timestamp = 0
        for i in range(nb_max_stored_orders):
            time_stamp = time.time()
            test_inst.upsert_order(i, {"id": i,
                                       "timestamp": time_stamp,
                                       "status": OrderStatus.CLOSED.value})
            if i == nb_max_stored_orders/2:
                max_timestamp = time_stamp
            time.sleep(0.000000001)

        assert len(test_inst.orders) == nb_max_stored_orders
        test_inst.upsert_order(nb_max_stored_orders+1, {"id": nb_max_stored_orders+1,
                                                        "timestamp": time.time(),
                                                        "status": OrderStatus.CLOSED.value})

        assert not [order for order in test_inst.orders.values() if order["timestamp"] < max_timestamp]

        assert len(test_inst.orders) == nb_max_stored_orders/2+1

    @staticmethod
    def create_fake_order(o_id, status, symbol, timestamp):
        return {
            "id": o_id,
            "status": status,
            "symbol": symbol,
            "timestamp": timestamp
        }

    def test_select_orders(self):
        test_inst = ExchangePersonalData()

        symbol_1 = "BTC/USDT"
        symbol_2 = "ETH/BTC"
        symbol_3 = "ETH/USDT"

        # without orders
        test_inst.orders = {}
        assert test_inst.get_all_orders(symbol_1, None, None) == []
        assert test_inst.get_open_orders(symbol_2, None, None) == []
        assert test_inst.get_closed_orders(symbol_3, None, None) == []

        order_1 = self.create_fake_order(10, OrderStatus.CLOSED.value, symbol_1, 0)
        test_inst.upsert_order(10, order_1)
        assert test_inst.get_all_orders(symbol_1, None, None) == [order_1]
        assert test_inst.get_open_orders(symbol_1, None, None) == []
        assert test_inst.get_closed_orders(symbol_1, None, None) == [order_1]

        order_2 = self.create_fake_order(100, OrderStatus.OPEN.value, symbol_3, 0)
        test_inst.upsert_order(100, order_2)
        assert test_inst.get_all_orders(symbol_3, None, None) == [order_2]
        assert test_inst.get_open_orders(symbol_3, None, None) == [order_2]
        assert test_inst.get_closed_orders(symbol_3, None, None) == []
        assert test_inst.get_all_orders(symbol_2, None, None) == []
        assert test_inst.get_open_orders(symbol_2, None, None) == []
        assert test_inst.get_closed_orders(symbol_2, None, None) == []
        assert test_inst.get_all_orders(None, None, None) == [order_1, order_2]
        assert test_inst.get_open_orders(None, None, None) == [order_2]
        assert test_inst.get_closed_orders(None, None, None) == [order_1]

        order_3 = self.create_fake_order(10, OrderStatus.OPEN.value, symbol_2, 0)
        test_inst.upsert_order(10, order_3)
        assert test_inst.get_all_orders(symbol_2, None, None) == [order_3]
        assert test_inst.get_open_orders(symbol_2, None, None) == [order_3]
        assert test_inst.get_closed_orders(symbol_2, None, None) == []
        assert test_inst.get_all_orders(None, None, None) == [order_3, order_2]
        assert test_inst.get_open_orders(None, None, None) == [order_3, order_2]
        assert test_inst.get_closed_orders(None, None, None) == []
        assert test_inst.get_all_orders(symbol_1, None, None) == []
        assert test_inst.get_open_orders(symbol_1, None, None) == []
        assert test_inst.get_closed_orders(symbol_1, None, None) == []
        assert test_inst.get_all_orders(symbol_3, None, None) == [order_2]
        assert test_inst.get_open_orders(symbol_3, None, None) == [order_2]
        assert test_inst.get_closed_orders(symbol_3, None, None) == []

        order_4 = self.create_fake_order(11, OrderStatus.OPEN.value, symbol_1, 100)
        order_5 = self.create_fake_order(12, OrderStatus.CLOSED.value, symbol_2, 1000)
        order_6 = self.create_fake_order(13, OrderStatus.CLOSED.value, symbol_3, 50)
        test_inst.upsert_orders([order_4, order_5, order_6])
        assert test_inst.get_all_orders(symbol_1, None, None) == [order_4]
        assert test_inst.get_open_orders(symbol_1, None, None) == [order_4]
        assert test_inst.get_closed_orders(symbol_1, None, None) == []
        assert test_inst.get_all_orders(symbol_2, None, None) == [order_3, order_5]
        assert test_inst.get_open_orders(symbol_2, None, None) == [order_3]
        assert test_inst.get_closed_orders(symbol_2, None, None) == [order_5]
        assert test_inst.get_all_orders(symbol_3, None, None) == [order_2, order_6]
        assert test_inst.get_open_orders(symbol_3, None, None) == [order_2]
        assert test_inst.get_closed_orders(symbol_3, None, None) == [order_6]
        assert test_inst.get_all_orders(None, None, None) == [order_3, order_2, order_4, order_5, order_6]
        assert test_inst.get_open_orders(None, None, None) == [order_3, order_2, order_4]
        assert test_inst.get_closed_orders(None, None, None) == [order_5, order_6]

        # test limit
        assert test_inst.get_all_orders(None, None, 3) == [order_3, order_2, order_4]
        assert test_inst.get_open_orders(None, None, 3) == [order_3, order_2, order_4]
        assert test_inst.get_closed_orders(None, None, 3) == [order_5, order_6]

        # test timestamps
        assert test_inst.get_all_orders(None, 30, 3) == [order_3, order_2]
        assert test_inst.get_open_orders(None, 1000, None) == [order_3, order_2, order_4]
        assert test_inst.get_closed_orders(None, 60, 1) == [order_6]



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import ccxt
import pytest
from octobot_trading.exchanges import ExchangeSimulator

from tests.test_utils.config import load_test_config
from octobot_trading.exchanges import ExchangeManager

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


class TestExchangeManager:
    @staticmethod
    async def init_default():
        config = load_test_config()
        exchange_manager = ExchangeManager(config, ccxt.binance, is_simulated=True)
        await exchange_manager.initialize()
        return config, exchange_manager

    async def test_create_exchanges(self):
        _, exchange_manager = await self.init_default()

        assert exchange_manager.exchange_dispatcher is not None
        assert exchange_manager.exchange is not None
        assert exchange_manager.is_ready
        assert exchange_manager.client_symbols is not None

        # simulated
        exchange_manager.is_simulated = True
        await exchange_manager.create_exchanges()

        assert isinstance(exchange_manager.exchange, ExchangeSimulator)

    async def test_enabled(self):
        _, exchange_manager = await self.init_default()

        assert exchange_manager.enabled()

    async def test__set_config_traded_pairs(self):
        _, exchange_manager = await self.init_default()

        exchange_manager.config[CONFIG_CRYPTO_CURRENCIES] = {
            "Neo": {
                "pairs": ["NEO/BTC"]
            },
            "Ethereum": {
                "pairs": ["ETH/USDT"]
            },
            "Icon": {
                "pairs": ["ICX/BTC"]
            }
        }

        exchange_manager._set_config_traded_pairs()

        assert exchange_manager.cryptocurrencies_traded_pairs == {
            "Ethereum": ["ETH/USDT"],
            "Icon": ["ICX/BTC"],
            "Neo": ["NEO/BTC"]
        }

    async def test__set_config_traded_pairs_wildcard(self):
        _, exchange_manager = await self.init_default()
        exchange_manager.config[CONFIG_CRYPTO_CURRENCIES] = {
            "Bitcoin": {
                "pairs": "*",
                "quote": "BTC"
            }
        }

        exchange_manager._set_config_traded_pairs()

        assert "ICX/BTC" in exchange_manager.cryptocurrencies_traded_pairs["Bitcoin"]
        assert "NEO/BTC" in exchange_manager.cryptocurrencies_traded_pairs["Bitcoin"]
        assert "VEN/BTC" in exchange_manager.cryptocurrencies_traded_pairs["Bitcoin"]
        assert "XLM/BTC" in exchange_manager.cryptocurrencies_traded_pairs["Bitcoin"]
        assert "ONT/BTC" in exchange_manager.cryptocurrencies_traded_pairs["Bitcoin"]
        assert "BTC/USDT" not in exchange_manager.cryptocurrencies_traded_pairs["Bitcoin"]
        assert "ETH/USDT" not in exchange_manager.cryptocurrencies_traded_pairs["Bitcoin"]
        assert "NEO/BNB" not in exchange_manager.cryptocurrencies_traded_pairs["Bitcoin"]

    async def test__set_config_traded_pairs_wildcard_with_add(self):
        _, exchange_manager = await self.init_default()
        exchange_manager.config[CONFIG_CRYPTO_CURRENCIES] = {
            "Bitcoin": {
                "pairs": "*",
                "quote": "BTC",
                "add": ["BTC/USDT"]
            }
        }

        exchange_manager._set_config_traded_pairs()

        assert "ICX/BTC" in exchange_manager.cryptocurrencies_traded_pairs["Bitcoin"]
        assert "NEO/BTC" in exchange_manager.cryptocurrencies_traded_pairs["Bitcoin"]
        assert "VEN/BTC" in exchange_manager.cryptocurrencies_traded_pairs["Bitcoin"]
        assert "XLM/BTC" in exchange_manager.cryptocurrencies_traded_pairs["Bitcoin"]
        assert "ONT/BTC" in exchange_manager.cryptocurrencies_traded_pairs["Bitcoin"]
        assert "BTC/USDT" in exchange_manager.cryptocurrencies_traded_pairs["Bitcoin"]
        assert "ETH/USDT" not in exchange_manager.cryptocurrencies_traded_pairs["Bitcoin"]
        assert "NEO/BNB" not in exchange_manager.cryptocurrencies_traded_pairs["Bitcoin"]



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import platform
import threading
from asyncio import CancelledError

from octobot.constants import FORCE_ASYNCIO_DEBUG_OPTION
from octobot_interfaces.api.interfaces import stop_interfaces
from octobot_commons.asyncio_tools import run_coroutine_in_asyncio_loop
from octobot_commons.logging.logging_util import get_logger
from octobot_services.api.service_feeds import stop_service_feed


class TaskManager:
    """TaskManager class:
        - Create, manage and stop octobot tasks
    """

    def __init__(self, octobot):
        self.octobot = octobot

        # Logger
        self.logger = get_logger(self.get_name())

        self.async_loop = None
        self.ready = False
        self.watcher = None
        self.tools_task_group = None
        self.current_loop_thread = None

    def init_async_loop(self):
        self.async_loop = asyncio.get_running_loop()

    async def start_tools_tasks(self, run_in_new_thread=False):
        task_list = []
        # if self.octobot.initializer.performance_analyser:
        #     task_list.append(self.octobot.initializer.performance_analyser.start_monitoring())

        if self.octobot.community_handler:
            task_list.append(self.octobot.community_handler.start_community_task())

        self.octobot.async_loop = self.async_loop
        self.ready = True
        self.tools_task_group = asyncio.gather(*task_list)

        # if run_in_new_thread:
        #     self._create_new_asyncio_main_loop()
        # else:
        #     self.current_loop_thread = threading.current_thread()
        #     await self.tools_task_group

    async def join_tasks(self):
        try:
            await asyncio.gather(*asyncio.all_tasks(asyncio.get_event_loop()))
        except CancelledError:
            self.logger.error("CancelledError raised")

    def stop_tasks(self):
        stop_coroutines = []
        # Notify stopping
        # if self.octobot.config[CONFIG_NOTIFICATION_INSTANCE].enabled(CONFIG_NOTIFICATION_GLOBAL_INFO):
        #     # To be improved with a full async implementation
        #     # To be done : "asyncio.run" --> replaced by a simple await
        #     # PR discussion : https://github.com/Drakkar-Software/OctoBot/pull/563#discussion_r248088266
        #     stop_coroutines.append(self.octobot.config[CONFIG_NOTIFICATION_INSTANCE]
        #                            .notify_with_all(NOTIFICATION_STOPPING_MESSAGE))

        self.logger.info("Stopping threads ...")

        # stop interfaces
        stop_coroutines.append(stop_interfaces(self.octobot.interface_factory.interface_list))

        # stop service feeds
        for service_feed in self.octobot.service_feed_factory.service_feeds:
            stop_coroutines.append(stop_service_feed(service_feed))

        if self.tools_task_group:
            self.async_loop.call_soon_threadsafe(self.tools_task_group.cancel)

        # close community session
        # stop_coroutines.append(self.octobot.metrics_handler.stop_task())

        # TODO: handle proper stop
        # for task in asyncio.all_tasks(self.async_loop):
        #     task.cancel()
        #
        # self.async_loop.close()
        # loop = asyncio.new_event_loop()
        # asyncio.set_event_loop(loop)
        
        self.async_loop.call_soon_threadsafe(asyncio.gather(*stop_coroutines, loop=self.async_loop))

        self.logger.info("Threads stopped.")

    @classmethod
    def get_name(cls):
        return cls.__name__

    def _create_new_asyncio_main_loop(self):
        self.async_loop = asyncio.new_event_loop()
        self.async_loop.set_debug(FORCE_ASYNCIO_DEBUG_OPTION)
        asyncio.set_event_loop(self.async_loop)
        self.current_loop_thread = threading.Thread(target=self.async_loop.run_forever,
                                                    name=f"{self.get_name()} new asyncio main loop")
        self.current_loop_thread.start()

    def run_in_main_asyncio_loop(self, coroutine):
        # restart a new loop if necessary (for backtesting analysis)
        # if backtesting_enabled(self.octobot.config) and self.async_loop.is_closed(): TODO
        #     self.logger.debug("Main loop is closed, starting a new main loop.")
        #     self._create_new_asyncio_main_loop()

        return run_coroutine_in_asyncio_loop(coroutine, self.async_loop)



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import ccxt
import pytest

from config import TraderOrderType, CONFIG_SIMULATOR, CONFIG_SIMULATOR_FEES, CONFIG_SIMULATOR_FEES_MAKER, \
    CONFIG_SIMULATOR_FEES_TAKER
from tests.test_utils.config import load_test_config
from tests.test_utils.order_util import fill_limit_or_stop_order, fill_market_order
from octobot_trading.exchanges import ExchangeManager
from trading.trader.order import BuyMarketOrder, SellLimitOrder, BuyLimitOrder, SellMarketOrder, \
    StopLossOrder
from trading.trader.portfolio import Portfolio
from trading.trader.trader_simulator import TraderSimulator


# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


class TestPortfolio:
    @staticmethod
    async def init_default():
        config = load_test_config()
        exchange_manager = ExchangeManager(config, ccxt.binance, is_simulated=True)
        await exchange_manager.initialize()
        exchange_inst = exchange_manager.get_exchange()
        trader_inst = TraderSimulator(config, exchange_inst, 1)
        portfolio_inst = Portfolio(config, trader_inst)
        await portfolio_inst.initialize()
        trader_inst.stop_order_manager()
        return config, portfolio_inst, exchange_inst, trader_inst

    async def test_load_portfolio(self):
        _, portfolio_inst, _, _ = await self.init_default()
        await portfolio_inst._load_portfolio()
        assert portfolio_inst.portfolio == {
            'BTC': {'available': 10, 'total': 10},
            'USD': {'available': 1000, 'total': 1000}
        }

    async def test_get_portfolio_from_amount_dict(self):
        assert Portfolio.get_portfolio_from_amount_dict({"zyx": 10, "BTC": 1}) == {
            'zyx': {'available': 10, 'total': 10},
            'BTC': {'available': 1, 'total': 1}
        }
        assert Portfolio.get_portfolio_from_amount_dict({}) == {}
        with pytest.raises(RuntimeError):
            Portfolio.get_portfolio_from_amount_dict({"zyx": "10", "BTC": 1})

    async def test_get_currency_portfolio(self):
        _, portfolio_inst, _, _ = await self.init_default()
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == 10
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.TOTAL) == 10
        assert portfolio_inst.get_currency_portfolio("NANO", Portfolio.TOTAL) == 0

    async def test_update_portfolio_data(self):
        _, portfolio_inst, _, _ = await self.init_default()
        portfolio_inst._update_portfolio_data("BTC", -5)
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.TOTAL) == 5
        portfolio_inst._update_portfolio_data("BTC", -6, total=False, available=True)
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == 4
        portfolio_inst._update_portfolio_data("XRP", 4.5, total=True, available=True)
        assert portfolio_inst.get_currency_portfolio("XRP", Portfolio.AVAILABLE) == 4.5

    async def test_update_portfolio_available(self):
        _, portfolio_inst, _, trader_inst = await self.init_default()

        # Test buy order
        market_buy = BuyMarketOrder(trader_inst)
        market_buy.new(TraderOrderType.BUY_MARKET,
                       "BTC/USD",
                       70,
                       10,
                       70)
        # test buy order creation
        portfolio_inst.update_portfolio_available(market_buy, True)
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == 10
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.AVAILABLE) == 300
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.TOTAL) == 10
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.TOTAL) == 1000

        # test buy order canceled --> return to init state and the update_portfolio will sync TOTAL with AVAILABLE
        portfolio_inst.update_portfolio_available(market_buy, False)
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == 10
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.AVAILABLE) == 1000
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.TOTAL) == 10
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.TOTAL) == 1000

        # Test sell order
        limit_sell = SellLimitOrder(trader_inst)
        limit_sell.new(TraderOrderType.SELL_LIMIT,
                       "BTC/USD",
                       60,
                       8,
                       60)
        # test sell order creation
        portfolio_inst.update_portfolio_available(limit_sell, True)
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == 2
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.AVAILABLE) == 1000
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.TOTAL) == 10
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.TOTAL) == 1000

        # test sell order canceled --> return to init state and the update_portfolio will sync TOTAL with AVAILABLE
        portfolio_inst.update_portfolio_available(limit_sell, False)
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == 10
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.AVAILABLE) == 1000
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.TOTAL) == 10
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.TOTAL) == 1000

    async def test_update_portfolio(self):
        _, portfolio_inst, _, trader_inst = await self.init_default()

        # Test buy order
        limit_buy = BuyLimitOrder(trader_inst)
        limit_buy.new(TraderOrderType.BUY_LIMIT,
                      "BTC/USD",
                      70,
                      10,
                      70)

        # update portfolio with creations
        portfolio_inst.update_portfolio_available(limit_buy, True)
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == 10
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.AVAILABLE) == 300

        await fill_limit_or_stop_order(limit_buy, 69, 71)

        await portfolio_inst.update_portfolio(limit_buy)
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == 20
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.AVAILABLE) == 300
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.TOTAL) == 20
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.TOTAL) == 300

        # Test buy order
        market_sell = SellMarketOrder(trader_inst)
        market_sell.new(TraderOrderType.SELL_MARKET,
                        "BTC/USD",
                        80,
                        8,
                        80)

        # update portfolio with creations
        portfolio_inst.update_portfolio_available(market_sell, True)
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == 12
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.AVAILABLE) == 300

        await fill_market_order(market_sell, 80)

        # when filling market sell
        await portfolio_inst.update_portfolio(market_sell)
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == 12
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.AVAILABLE) == 940
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.TOTAL) == 12
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.TOTAL) == 940

    async def test_update_portfolio_with_filled_orders(self):
        _, portfolio_inst, exchange_inst, trader_inst = await self.init_default()

        # force fees => should have consequences
        exchange_inst.config[CONFIG_SIMULATOR][CONFIG_SIMULATOR_FEES] = {
            CONFIG_SIMULATOR_FEES_MAKER: 0.05,
            CONFIG_SIMULATOR_FEES_TAKER: 0.1
        }

        # Test buy order
        market_sell = SellMarketOrder(trader_inst)
        market_sell.new(TraderOrderType.SELL_MARKET,
                        "BTC/USD",
                        70,
                        3,
                        70)

        await fill_market_order(market_sell, 70)

        # Test sell order
        limit_sell = SellLimitOrder(trader_inst)
        limit_sell.new(TraderOrderType.SELL_LIMIT,
                       "BTC/USD",
                       100,
                       4.2,
                       100)

        # Test stop loss order
        stop_loss = StopLossOrder(trader_inst)
        stop_loss.new(TraderOrderType.STOP_LOSS,
                      "BTC/USD",
                      80,
                      4.2,
                      80)

        limit_sell.add_linked_order(stop_loss)
        stop_loss.add_linked_order(limit_sell)

        # Test buy order
        limit_buy = BuyLimitOrder(trader_inst)
        limit_buy.new(TraderOrderType.BUY_LIMIT,
                      "BTC/USD",
                      50,
                      2,
                      50)

        await fill_limit_or_stop_order(limit_buy, 49, 51)

        # update portfolio with creations
        portfolio_inst.update_portfolio_available(market_sell, True)
        portfolio_inst.update_portfolio_available(limit_sell, True)
        portfolio_inst.update_portfolio_available(stop_loss, True)
        portfolio_inst.update_portfolio_available(limit_buy, True)

        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == 2.8
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.AVAILABLE) == 900

        # when cancelling limit sell, market sell and stop orders
        portfolio_inst.update_portfolio_available(stop_loss, False)
        portfolio_inst.update_portfolio_available(limit_sell, False)

        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == 7
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.AVAILABLE) == 900

        # when filling limit buy
        await portfolio_inst.update_portfolio(limit_buy)

        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == 8.999
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.AVAILABLE) == 900
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.TOTAL) == 11.999
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.TOTAL) == 900

        # when filling market sell
        await portfolio_inst.update_portfolio(market_sell)

        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == 8.999
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.AVAILABLE) == 1109.79
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.TOTAL) == 8.999
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.TOTAL) == 1109.79

    async def test_update_portfolio_with_cancelled_orders(self):
        _, portfolio_inst, exchange_inst, trader_inst = await self.init_default()

        # force fees => shouldn't do anything
        exchange_inst.config[CONFIG_SIMULATOR][CONFIG_SIMULATOR_FEES] = {
            CONFIG_SIMULATOR_FEES_MAKER: 0.05,
            CONFIG_SIMULATOR_FEES_TAKER: 0.1
        }

        # Test buy order
        market_sell = SellMarketOrder(trader_inst)
        market_sell.new(TraderOrderType.SELL_MARKET,
                        "BTC/USD",
                        80,
                        4.1,
                        80)

        # Test sell order
        limit_sell = SellLimitOrder(trader_inst)
        limit_sell.new(TraderOrderType.SELL_LIMIT,
                       "BTC/USD",
                       10,
                       4.2,
                       10)

        # Test stop loss order
        stop_loss = StopLossOrder(trader_inst)
        stop_loss.new(TraderOrderType.STOP_LOSS,
                      "BTC/USD",
                      80,
                      3.6,
                      80)

        portfolio_inst.update_portfolio_available(stop_loss, True)
        portfolio_inst.update_portfolio_available(limit_sell, True)

        # Test buy order
        limit_buy = BuyLimitOrder(trader_inst)
        limit_buy.new(TraderOrderType.BUY_LIMIT,
                      "BTC/USD",
                      50,
                      4,
                      50)

        portfolio_inst.update_portfolio_available(limit_buy, True)
        portfolio_inst.update_portfolio_available(market_sell, True)

        assert round(portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE), 1) == 1.7
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.AVAILABLE) == 800
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.TOTAL) == 10
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.TOTAL) == 1000

        # with no filled orders
        portfolio_inst.update_portfolio_available(stop_loss, False)
        portfolio_inst.update_portfolio_available(limit_sell, False)
        portfolio_inst.update_portfolio_available(limit_buy, False)
        portfolio_inst.update_portfolio_available(market_sell, False)

        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == 10
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.AVAILABLE) == 1000
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.TOTAL) == 10
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.TOTAL) == 1000

    async def test_update_portfolio_with_stop_loss_orders(self):
        _, portfolio_inst, _, trader_inst = await self.init_default()

        # Test buy order
        limit_sell = SellLimitOrder(trader_inst)
        limit_sell.new(TraderOrderType.SELL_LIMIT,
                       "BTC/USD",
                       90,
                       4,
                       90)

        # Test buy order
        limit_buy = BuyLimitOrder(trader_inst)
        limit_buy.new(TraderOrderType.BUY_LIMIT,
                      "BTC/USD",
                      50,
                      4,
                      50)

        # Test stop loss order
        stop_loss = StopLossOrder(trader_inst)
        stop_loss.new(TraderOrderType.STOP_LOSS,
                      "BTC/USD",
                      60,
                      4,
                      60)

        await fill_limit_or_stop_order(stop_loss, 59, 61)

        portfolio_inst.update_portfolio_available(stop_loss, True)
        portfolio_inst.update_portfolio_available(limit_sell, True)
        portfolio_inst.update_portfolio_available(limit_buy, True)

        assert round(portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE), 1) == 6
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.AVAILABLE) == 800
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.TOTAL) == 10
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.TOTAL) == 1000

        # cancel limits
        portfolio_inst.update_portfolio_available(limit_buy, False)
        portfolio_inst.update_portfolio_available(limit_sell, False)

        # filling stop loss
        # typical stop loss behavior --> update available before update portfolio
        await portfolio_inst.update_portfolio(stop_loss)

        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == 6
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.AVAILABLE) == 1240
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.TOTAL) == 6
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.TOTAL) == 1240

    async def test_update_portfolio_with_some_filled_orders(self):
        _, portfolio_inst, _, trader_inst = await self.init_default()

        # Test buy order
        limit_sell = SellLimitOrder(trader_inst)
        limit_sell.new(TraderOrderType.SELL_LIMIT,
                       "BTC/USD",
                       90,
                       4,
                       90)

        # Test buy order
        limit_buy = BuyLimitOrder(trader_inst)
        limit_buy.new(TraderOrderType.BUY_LIMIT,
                      "BTC/USD",
                      60,
                      2,
                      60)

        # Test buy order
        limit_buy_2 = BuyLimitOrder(trader_inst)
        limit_buy_2.new(TraderOrderType.BUY_LIMIT,
                        "BTC/USD",
                        50,
                        4,
                        50)

        # Test sell order
        limit_sell_2 = SellLimitOrder(trader_inst)
        limit_sell_2.new(TraderOrderType.SELL_LIMIT,
                         "BTC/USD",
                         10,
                         2,
                         10)

        # Test stop loss order
        stop_loss_2 = StopLossOrder(trader_inst)
        stop_loss_2.new(TraderOrderType.STOP_LOSS,
                        "BTC/USD",
                        10,
                        2,
                        10)

        # Test sell order
        limit_sell_3 = SellLimitOrder(trader_inst)
        limit_sell_3.new(TraderOrderType.SELL_LIMIT,
                         "BTC/USD",
                         20,
                         1,
                         20)

        # Test stop loss order
        stop_loss_3 = StopLossOrder(trader_inst)
        stop_loss_3.new(TraderOrderType.STOP_LOSS,
                        "BTC/USD",
                        20,
                        1,
                        20)

        portfolio_inst.update_portfolio_available(stop_loss_2, True)
        portfolio_inst.update_portfolio_available(stop_loss_3, True)
        portfolio_inst.update_portfolio_available(limit_sell, True)
        portfolio_inst.update_portfolio_available(limit_sell_2, True)
        portfolio_inst.update_portfolio_available(limit_sell_3, True)
        portfolio_inst.update_portfolio_available(limit_buy, True)
        portfolio_inst.update_portfolio_available(limit_buy_2, True)

        assert round(portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE), 1) == 3
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.AVAILABLE) == 680
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.TOTAL) == 10
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.TOTAL) == 1000

        # cancels
        portfolio_inst.update_portfolio_available(stop_loss_3, False)
        portfolio_inst.update_portfolio_available(limit_sell_2, False)
        portfolio_inst.update_portfolio_available(limit_buy, False)

        # filling
        await fill_limit_or_stop_order(stop_loss_2, 9, 11)
        await fill_limit_or_stop_order(limit_sell, 89, 91)
        await fill_limit_or_stop_order(limit_sell_3, 19, 21)
        await fill_limit_or_stop_order(limit_buy_2, 49, 51)

        await portfolio_inst.update_portfolio(stop_loss_2)
        await portfolio_inst.update_portfolio(limit_sell)
        await portfolio_inst.update_portfolio(limit_sell_3)
        await portfolio_inst.update_portfolio(limit_buy_2)

        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == 7
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.AVAILABLE) == 1200
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.TOTAL) == 7
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.TOTAL) == 1200

    async def test_update_portfolio_with_multiple_filled_orders(self):
        _, portfolio_inst, _, trader_inst = await self.init_default()

        # Test buy order
        limit_sell = SellLimitOrder(trader_inst)
        limit_sell.new(TraderOrderType.SELL_LIMIT,
                       "BTC/USD",
                       90,
                       4,
                       90)

        # Test buy order
        limit_buy = BuyLimitOrder(trader_inst)
        limit_buy.new(TraderOrderType.BUY_LIMIT,
                      "BTC/USD",
                      60,
                      2,
                      60)

        # Test buy order
        limit_buy_2 = BuyLimitOrder(trader_inst)
        limit_buy_2.new(TraderOrderType.BUY_LIMIT,
                        "BTC/USD",
                        50,
                        4,
                        50)

        # Test buy order
        limit_buy_3 = BuyLimitOrder(trader_inst)
        limit_buy_3.new(TraderOrderType.BUY_LIMIT,
                        "BTC/USD",
                        46,
                        2,
                        46)

        # Test buy order
        limit_buy_4 = BuyLimitOrder(trader_inst)
        limit_buy_4.new(TraderOrderType.BUY_LIMIT,
                        "BTC/USD",
                        41,
                        1.78,
                        41)

        # Test buy order
        limit_buy_5 = BuyLimitOrder(trader_inst)
        limit_buy_5.new(TraderOrderType.BUY_LIMIT,
                        "BTC/USD",
                        0.2122427,
                        3.72448,
                        0.2122427)

        # Test buy order
        limit_buy_6 = BuyLimitOrder(trader_inst)
        limit_buy_6.new(TraderOrderType.BUY_LIMIT,
                        "BTC/USD",
                        430,
                        1.05,
                        430)

        # Test sell order
        limit_sell_2 = SellLimitOrder(trader_inst)
        limit_sell_2.new(TraderOrderType.SELL_LIMIT,
                         "BTC/USD",
                         10,
                         2,
                         10)

        # Test stop loss order
        stop_loss_2 = StopLossOrder(trader_inst)
        stop_loss_2.new(TraderOrderType.STOP_LOSS,
                        "BTC/USD",
                        10,
                        2,
                        10)

        # Test sell order
        limit_sell_3 = SellLimitOrder(trader_inst)
        limit_sell_3.new(TraderOrderType.SELL_LIMIT,
                         "BTC/USD",
                         20,
                         1,
                         20)

        # Test stop loss order
        stop_loss_3 = StopLossOrder(trader_inst)
        stop_loss_3.new(TraderOrderType.STOP_LOSS,
                        "BTC/USD",
                        20,
                        1,
                        20)

        # Test sell order
        limit_sell_4 = SellLimitOrder(trader_inst)
        limit_sell_4.new(TraderOrderType.SELL_LIMIT,
                         "BTC/USD",
                         50,
                         0.2,
                         50)

        # Test stop loss order
        stop_loss_4 = StopLossOrder(trader_inst)
        stop_loss_4.new(TraderOrderType.STOP_LOSS,
                        "BTC/USD",
                        45,
                        0.2,
                        45)

        # Test sell order
        limit_sell_5 = SellLimitOrder(trader_inst)
        limit_sell_5.new(TraderOrderType.SELL_LIMIT,
                         "BTC/USD",
                         11,
                         0.7,
                         11)

        # Test stop loss order
        stop_loss_5 = StopLossOrder(trader_inst)
        stop_loss_5.new(TraderOrderType.STOP_LOSS,
                        "BTC/USD",
                        9,
                        0.7,
                        9)

        portfolio_inst.update_portfolio_available(stop_loss_2, True)
        portfolio_inst.update_portfolio_available(stop_loss_3, True)
        portfolio_inst.update_portfolio_available(stop_loss_4, True)
        portfolio_inst.update_portfolio_available(stop_loss_5, True)
        portfolio_inst.update_portfolio_available(limit_sell, True)
        portfolio_inst.update_portfolio_available(limit_sell_2, True)
        portfolio_inst.update_portfolio_available(limit_sell_3, True)
        portfolio_inst.update_portfolio_available(limit_sell_4, True)
        portfolio_inst.update_portfolio_available(limit_sell_5, True)
        portfolio_inst.update_portfolio_available(limit_buy, True)
        portfolio_inst.update_portfolio_available(limit_buy_2, True)
        portfolio_inst.update_portfolio_available(limit_buy_3, True)
        portfolio_inst.update_portfolio_available(limit_buy_4, True)
        portfolio_inst.update_portfolio_available(limit_buy_5, True)
        portfolio_inst.update_portfolio_available(limit_buy_6, True)

        assert round(portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE), 1) == 2.1
        assert round(portfolio_inst.get_currency_portfolio("USD", Portfolio.AVAILABLE), 7) == 62.7295063
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.TOTAL) == 10
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.TOTAL) == 1000

        # cancels
        portfolio_inst.update_portfolio_available(stop_loss_3, False)
        portfolio_inst.update_portfolio_available(stop_loss_5, False)
        portfolio_inst.update_portfolio_available(limit_sell_2, False)
        portfolio_inst.update_portfolio_available(limit_buy, False)
        portfolio_inst.update_portfolio_available(limit_buy_3, False)
        portfolio_inst.update_portfolio_available(limit_buy_5, False)
        portfolio_inst.update_portfolio_available(limit_sell_4, False)

        # filling
        await fill_limit_or_stop_order(stop_loss_2, 9, 11)
        await fill_limit_or_stop_order(limit_sell, 89, 91)
        await fill_limit_or_stop_order(limit_sell_3, 19, 21)
        await fill_limit_or_stop_order(limit_buy_2, 49, 51)
        await fill_limit_or_stop_order(limit_sell_5, 9, 12)
        await fill_limit_or_stop_order(stop_loss_4, 44, 46)
        await fill_limit_or_stop_order(limit_buy_4, 40, 42)
        await fill_limit_or_stop_order(limit_buy_5, 0.2122426, 0.2122428)
        await fill_limit_or_stop_order(limit_buy_6, 429, 431)

        await portfolio_inst.update_portfolio(stop_loss_2)
        await portfolio_inst.update_portfolio(limit_buy_4)
        await portfolio_inst.update_portfolio(limit_sell)
        await portfolio_inst.update_portfolio(limit_sell_3)
        await portfolio_inst.update_portfolio(limit_buy_2)
        await portfolio_inst.update_portfolio(limit_sell_5)
        await portfolio_inst.update_portfolio(stop_loss_4)
        await portfolio_inst.update_portfolio(limit_buy_5)
        await portfolio_inst.update_portfolio(limit_buy_6)

        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == 12.65448
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.AVAILABLE) == 692.22
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.TOTAL) == 12.65448
        assert round(portfolio_inst.get_currency_portfolio("USD", Portfolio.TOTAL), 7) == 691.4295063

    async def test_update_portfolio_with_multiple_symbols_orders(self):
        _, portfolio_inst, _, trader_inst = await self.init_default()

        # Test buy order
        market_buy = BuyMarketOrder(trader_inst)
        market_buy.new(TraderOrderType.BUY_MARKET,
                       "ETH/USD",
                       7,
                       100,
                       7)

        # test buy order creation
        portfolio_inst.update_portfolio_available(market_buy, True)
        assert portfolio_inst.get_currency_portfolio("ETH", Portfolio.AVAILABLE) == 0
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.AVAILABLE) == 300
        assert portfolio_inst.get_currency_portfolio("ETH", Portfolio.TOTAL) == 0
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.TOTAL) == 1000

        await fill_market_order(market_buy, 7)

        await portfolio_inst.update_portfolio(market_buy)
        assert portfolio_inst.get_currency_portfolio("ETH", Portfolio.AVAILABLE) == 100
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.AVAILABLE) == 300
        assert portfolio_inst.get_currency_portfolio("ETH", Portfolio.TOTAL) == 100
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.TOTAL) == 300

        # Test buy order
        market_buy = BuyMarketOrder(trader_inst)
        market_buy.new(TraderOrderType.BUY_MARKET,
                       "LTC/BTC",
                       0.0135222,
                       100,
                       0.0135222)

        # test buy order creation
        portfolio_inst.update_portfolio_available(market_buy, True)
        assert portfolio_inst.get_currency_portfolio("LTC", Portfolio.AVAILABLE) == 0
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == 8.647780000000001
        assert portfolio_inst.get_currency_portfolio("LTC", Portfolio.TOTAL) == 0
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.TOTAL) == 10

        await fill_market_order(market_buy, 0.0135222)

        await portfolio_inst.update_portfolio(market_buy)
        assert portfolio_inst.get_currency_portfolio("LTC", Portfolio.AVAILABLE) == 100
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == 8.647780000000001
        assert portfolio_inst.get_currency_portfolio("LTC", Portfolio.TOTAL) == 100
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.TOTAL) == 8.647780000000001

        # Test buy order
        limit_buy = BuyLimitOrder(trader_inst)
        limit_buy.new(TraderOrderType.BUY_LIMIT,
                      "XRP/BTC",
                      0.00012232132312312,
                      3000.1214545,
                      0.00012232132312312)

        # test buy order creation
        portfolio_inst.update_portfolio_available(limit_buy, True)
        assert portfolio_inst.get_currency_portfolio("XRP", Portfolio.AVAILABLE) == 0
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == 8.280801174155501
        assert portfolio_inst.get_currency_portfolio("XRP", Portfolio.TOTAL) == 0
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.TOTAL) == 8.647780000000001

        # cancel
        portfolio_inst.update_portfolio_available(limit_buy, False)
        assert portfolio_inst.get_currency_portfolio("XRP", Portfolio.AVAILABLE) == 0
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == 8.647780000000001
        assert portfolio_inst.get_currency_portfolio("XRP", Portfolio.TOTAL) == 0
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.TOTAL) == 8.647780000000001

    async def test_reset_portfolio_available(self):
        _, portfolio_inst, _, trader_inst = await self.init_default()

        # Test buy order
        limit_sell = SellLimitOrder(trader_inst)
        limit_sell.new(TraderOrderType.SELL_LIMIT,
                       "BTC/USD",
                       90,
                       4,
                       90)

        portfolio_inst.update_portfolio_available(limit_sell, True)
        portfolio_inst.reset_portfolio_available()

        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == 10
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.AVAILABLE) == 1000
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.TOTAL) == 10
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.TOTAL) == 1000

        # Test sell order
        limit_sell = SellLimitOrder(trader_inst)
        limit_sell.new(TraderOrderType.SELL_LIMIT,
                       "BTC/USD",
                       90,
                       4,
                       90)

        portfolio_inst.update_portfolio_available(limit_sell, True)
        # Test buy order
        limit_buy = BuyLimitOrder(trader_inst)
        limit_buy.new(TraderOrderType.BUY_LIMIT,
                      "VEN/BTC",
                      0.5,
                      4,
                      0.5)

        portfolio_inst.update_portfolio_available(limit_buy, True)

        # Test buy order
        btc_limit_buy = BuyLimitOrder(trader_inst)
        btc_limit_buy.new(TraderOrderType.BUY_LIMIT,
                          "BTC/USD",
                          10,
                          50,
                          10)

        portfolio_inst.update_portfolio_available(btc_limit_buy, True)

        # Test buy order
        btc_limit_buy2 = BuyLimitOrder(trader_inst)
        btc_limit_buy2.new(TraderOrderType.BUY_LIMIT,
                           "BTC/USD",
                           10,
                           50,
                           10)

        portfolio_inst.update_portfolio_available(btc_limit_buy2, True)

        # reset equivalent of the ven buy order
        portfolio_inst.reset_portfolio_available("BTC", 4*0.5)

        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == 6
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.TOTAL) == 10
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.AVAILABLE) == 0
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.TOTAL) == 1000

        # reset equivalent of the btc buy orders 1 and 2
        portfolio_inst.reset_portfolio_available("USD")

        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.AVAILABLE) == 6
        assert portfolio_inst.get_currency_portfolio("BTC", Portfolio.TOTAL) == 10
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.AVAILABLE) == 1000
        assert portfolio_inst.get_currency_portfolio("USD", Portfolio.TOTAL) == 1000



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import math

from evaluator.evaluator_matrix import EvaluatorMatrix
from tests.test_utils.config import load_test_config
from config import EvaluatorMatrixTypes, TimeFrames


def _get_tools():
    config = load_test_config()
    matrix_inst = EvaluatorMatrix(config)
    matrix = matrix_inst.matrix
    return matrix_inst, matrix, config


def test_init():
    matrix_inst, matrix, config = _get_tools()
    assert EvaluatorMatrixTypes.TA in matrix
    assert EvaluatorMatrixTypes.SOCIAL in matrix
    assert EvaluatorMatrixTypes.REAL_TIME in matrix
    assert EvaluatorMatrixTypes.STRATEGIES in matrix


def test_set_get_eval():
    time_frames = [None, TimeFrames.ONE_HOUR]
    values = [1, 1.02, math.pi, math.nan]
    for time_frame in time_frames:
        matrix_inst, matrix, config = _get_tools()
        for value in values:
            for matrix_type in EvaluatorMatrixTypes:
                key = f"{matrix_type}{value}"
                matrix_inst.set_eval(matrix_type, key, value, time_frame)
                if math.isnan(value):
                    assert matrix_inst.get_eval_note(matrix, matrix_type, key, time_frame) is None
                else:
                    assert matrix_inst.get_eval_note(matrix, matrix_type, key, time_frame) == value



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from abc import ABCMeta, ABC
import copy

from backtesting.backtesting_util import create_backtesting_config, create_backtesting_bot, \
    start_backtesting_bot, filter_wanted_symbols, update_starting_portfolio_if_required
from backtesting.abstract_backtesting_test import AbstractBacktestingTest
from config import CONFIG_EVALUATOR, CONFIG_BACKTESTING, CONFIG_BACKTESTING_DATA_FILES, CONFIG_TRADING_TENTACLES, \
    CONFIG_EVALUATORS_WILDCARD
from evaluator import Strategies
from evaluator.Strategies.strategies_evaluator import StrategiesEvaluator
from trading import modes
from trading.modes.abstract_trading_mode import AbstractTradingMode
from trading.modes import DailyTradingMode
from octobot_commons.tentacles_management.class_inspector import get_class_from_string, evaluator_parent_inspection, \
    trading_mode_parent_inspection
from backtesting.collector.data_file_manager import interpret_file_name, DATA_FILE_EXT
from tests.test_utils.config import load_test_config
from services.web_service import WebService


DEFAULT_SYMBOL = "ICX/BTC"
DATA_FILE_PATH = "tests/static/"


class AbstractStrategyTest(AbstractBacktestingTest, ABC):
    __metaclass__ = ABCMeta

    def initialize(self, strategy_evaluator_class, trading_mode_class=DailyTradingMode, config=None):
        if config is None:
            self.config = create_backtesting_config(load_test_config(), filter_symbols=False)
        else:
            self.config = config
        self.strategy_evaluator_class = strategy_evaluator_class
        self.trading_mode_class = trading_mode_class
        self._register_evaluators()
        self._register_only_strategy(strategy_evaluator_class)
        self._register_only_trading_mode(trading_mode_class)
        self._assert_init()

    def _assert_results(self, run_results, profitability, bot):
        # convenient for building tests
        # print(f"results: {run_results} expected: {profitability} => {run_results[0] >= profitability}")

        assert run_results[0] >= profitability

    async def _run_backtesting_with_current_config(self, symbol, data_file_to_use=None):
        config_to_use = copy.deepcopy(self.config)
        if data_file_to_use is not None:
            for index, datafile in enumerate(config_to_use[CONFIG_BACKTESTING][CONFIG_BACKTESTING_DATA_FILES]):
                _, file_symbol, _, _ = interpret_file_name(datafile)
                if symbol == file_symbol:
                    config_to_use[CONFIG_BACKTESTING][CONFIG_BACKTESTING_DATA_FILES][index] = \
                        DATA_FILE_PATH + data_file_to_use + DATA_FILE_EXT

        # do not activate web interface on standalone backtesting bot
        WebService.enable(config_to_use, False)
        filter_wanted_symbols(config_to_use, [symbol])
        update_starting_portfolio_if_required(config_to_use, symbol)
        bot = create_backtesting_bot(config_to_use)
        return await start_backtesting_bot(bot), bot

    def _register_only_strategy(self, strategy_evaluator_class):
        for evaluator_name in self.config[CONFIG_EVALUATOR]:
            if get_class_from_string(evaluator_name, StrategiesEvaluator, Strategies,
                                     evaluator_parent_inspection) is not None:
                self.config[CONFIG_EVALUATOR][evaluator_name] = False
        self.config[CONFIG_EVALUATOR][strategy_evaluator_class.get_name()] = True

    def _register_only_trading_mode(self, trading_mode_class):
        for trading_mode_name in self.config[CONFIG_TRADING_TENTACLES]:
            if get_class_from_string(trading_mode_name, AbstractTradingMode, modes,
                                     trading_mode_parent_inspection) is not None:
                self.config[CONFIG_TRADING_TENTACLES][trading_mode_name] = False
        self.config[CONFIG_TRADING_TENTACLES][trading_mode_class.get_name()] = True

    def _register_evaluators(self):
        required_evaluators = self.strategy_evaluator_class.get_required_evaluators(self.config)
        if required_evaluators != CONFIG_EVALUATORS_WILDCARD:
            for evaluator_name in self.config[CONFIG_EVALUATOR]:
                self.config[CONFIG_EVALUATOR][evaluator_name] = False
            for required_evaluator_name in required_evaluators:
                self.config[CONFIG_EVALUATOR][required_evaluator_name] = True

    def _assert_init(self):
        assert self.config
        assert self.config[CONFIG_EVALUATOR][self.strategy_evaluator_class.get_name()] is True
        assert self.config[CONFIG_TRADING_TENTACLES][self.trading_mode_class.get_name()] is True
